#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
from abc import abstractmethod
from collections import UserDict
from dataclasses import dataclass
from enum import Enum
from typing import (
    List,
    Optional,
    Sequence,
    Union,
)

from pydicom.sr.coding import Code
from pydicom.uid import UID

from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ContainerTypeCode,
    SpecimenSamplingProcedureCode,
)
from wsidicom.metadata.sample import (
    BaseSampling,
    Collection,
    Specimen,
    PreparationStep,
    Sample,
    Sampling,
    SamplingLocation,
    SlideSample,
    BaseSpecimen,
    SpecimenIdentifier,
    SampleLocalization,
    UnknownSampling,
)


class PreparationAction(Enum):
    SAMPLING = "sampling"
    COLLECTION = "collection"
    PROCESSING = "processing"
    STAINING = "staining"
    FIXATION = "fixation"
    EMBEDDING = "embedding"
    RECEIVING = "receiving"
    STORAGE = "storage"


@dataclass(frozen=True)
class SamplingConstraintJsonModel:
    """Simplified representation of a `Sampling` to use as sampling tree constraint,
    replacing the sampling with the identifier of the sampled specimen and the index of
    the sampling step within the step sequence of the specimen."""

    identifier: Union[str, SpecimenIdentifier]
    sampling_step_index: int

    @classmethod
    def to_json_model(cls, sampling: Sampling) -> "SamplingConstraintJsonModel":
        """Create json model for sampling."""
        return cls(
            identifier=sampling.specimen.identifier,
            sampling_step_index=sampling.specimen.samplings.index(sampling),
        )

    def from_json_model(
        self,
        specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen],
    ) -> Sampling:
        """Create sampling from json model.

        Parameters
        ----------
        specimens: UserDict[
            Union[str, SpecimenIdentifier], Specimen
        ]
            Dictionary providing specimens.

        Returns
        -------
        Sampling
            Created sampling.
        """
        specimen = specimens[self.identifier]
        return specimen.samplings[self.sampling_step_index]


@dataclass(frozen=True)
class SamplingJsonModel:
    """Simplified representation of a `Sampling`, replacing the sampled specimen with
    the idententifier and sampling constraints with simplified sampling constraints."""

    method: Optional[SpecimenSamplingProcedureCode] = None
    sampling_constraints: Optional[Sequence[SamplingConstraintJsonModel]] = None
    date_time: Optional[datetime.datetime] = None
    location: Optional[SamplingLocation] = None
    description: Optional[str] = None

    def from_json_model(
        self,
        specimen: BaseSpecimen,
        specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen],
    ) -> BaseSampling:
        """Create sampling from json model.

        Parameters
        ----------
        specimen: Specimen
            The specimen that has been sampled.
        specimens: UserDict[
            Union[str, SpecimenIdentifier], Specimen
        ]
            Dictionary providing specimens.

        Returns
        -------
        Sampling
            Created sampling.
        """
        if self.method is None or specimen.type is None:
            return UnknownSampling(
                specimen,
                self._get_sampling_constraints(specimens),
            )
        return Sampling(
            specimen,
            self.method,
            specimen.type,
            self._get_sampling_constraints(specimens),
            self.date_time,
            self.description,
            self.location,
        )

    def _get_sampling_constraints(
        self,
        specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen],
    ) -> Optional[List[Sampling]]:
        """
        Get list of constraint sampling this sampling.

        Parameters
        ---------
        specimens: UserDict[
            Union[str, SpecimenIdentifier], Specimen
        ]
            Dictionary providing specimens.

        Returns
        -------
        Optional[List[Sampling]]
            List of constraint samplings, or None if no constraints.
        """
        if self.sampling_constraints is None:
            return None
        return [
            constraint.from_json_model(specimens)
            for constraint in self.sampling_constraints
        ]


@dataclass(frozen=True)
class BaseSpecimenJsonModel:
    """Base json model for a specimen."""

    identifier: Union[str, SpecimenIdentifier]
    steps: List[Union[PreparationStep, SamplingJsonModel]]

    @abstractmethod
    def from_json_model(
        self, specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen]
    ) -> BaseSpecimen:
        """Return specimen created from this json model.

        Parameters
        ----------
        specimens: UserDict[
            Union[str, SpecimenIdentifier], Specimen
        ]
            Dictionary providing specimens.

        Returns
        -------
        Specimen
            Created specimen for this model.
        """
        raise NotImplementedError()


@dataclass(frozen=True)
class SpecimenJsonModel(BaseSpecimenJsonModel):
    type: AnatomicPathologySpecimenTypesCode
    container: Optional[ContainerTypeCode]

    def from_json_model(
        self,
        specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen],
    ) -> Specimen:
        extraction_step = next(
            (step for step in self.steps if isinstance(step, Collection)), None
        )
        specimen = Specimen(
            identifier=self.identifier,
            extraction_step=extraction_step,
            type=self.type,
            container=self.container,
        )
        for step in self.steps:
            if isinstance(step, SamplingJsonModel):
                step = step.from_json_model(specimen, specimens)
            elif isinstance(step, Collection):
                if step != extraction_step:
                    raise ValueError(
                        "There should be only one Collection step.",
                    )
                continue
            specimen.add(step)
        return specimen


@dataclass(frozen=True)
class SampleJsonModel(BaseSpecimenJsonModel):
    type: AnatomicPathologySpecimenTypesCode
    sampled_from: Sequence[SamplingConstraintJsonModel]
    container: Optional[ContainerTypeCode]

    def from_json_model(
        self,
        specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen],
    ) -> Sample:
        sample = Sample(
            identifier=self.identifier,
            type=self.type,
            sampled_from=[
                sample.from_json_model(specimens) for sample in self.sampled_from
            ],
            container=self.container,
        )
        for step in self.steps:
            if isinstance(step, SamplingJsonModel):
                step = step.from_json_model(sample, specimens)
            elif isinstance(step, Collection):
                raise ValueError(
                    "Collection step can only be added to an Specimen",
                )
            sample.add(step)
        return sample


@dataclass(frozen=True)
class SlideSampleJsonModel(BaseSpecimenJsonModel):
    anatomical_sites: Optional[Sequence[Code]] = None
    sampled_from: Optional[SamplingConstraintJsonModel] = None
    uid: Optional[UID] = None
    localization: Optional[SampleLocalization] = None
    short_description: Optional[str] = None
    detailed_description: Optional[str] = None

    def from_json_model(
        self,
        specimens: UserDict[Union[str, SpecimenIdentifier], BaseSpecimen],
    ) -> SlideSample:
        sample = SlideSample(
            identifier=self.identifier,
            anatomical_sites=self.anatomical_sites,
            sampled_from=(
                self.sampled_from.from_json_model(specimens)
                if self.sampled_from
                else None
            ),
            uid=self.uid,
            localization=self.localization,
            short_description=self.short_description,
            detailed_description=self.detailed_description,
        )
        for step in self.steps:
            if isinstance(step, SamplingJsonModel):
                raise ValueError("A SlideSample cannot be sampled to another specimen.")
            elif isinstance(step, Collection):
                raise ValueError(
                    "Collection step can only be added to an Specimen",
                )
            sample.add(step)
        return sample



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from http import HTTPStatus
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Tuple, Union

from dicomweb_client.api import DICOMfileClient, DICOMwebClient
from dicomweb_client.session_utils import create_session_from_auth
from pydicom import Dataset
from pydicom.uid import UID
from requests import HTTPError, Session
from requests.auth import AuthBase

from wsidicom.codec import determine_media_type
from wsidicom.uid import ANN_SOP_CLASS_UID, WSI_SOP_CLASS_UID

SOP_CLASS_UID = "00080016"
SOP_INSTANCE_UID = "00080018"
SERIES_INSTANCE_UID = "0020000E"
MODALITY = "00080060"
AVAILABLE_SOP_TRANSFER_SYNTAX_UID = "00083002"
WSI_MODALITY = "SM"
ANNOTATION_MODALITY = "ANN"


class WsiDicomWebClient:
    def __init__(self, client: Union[DICOMwebClient, DICOMfileClient]):
        """Create a WsiDicomWebClient.

        Parameters
        ----------
        client: Union[DICOMwebClient, DICOMfileClient]
            The client to use
        """
        self._client = client

    @classmethod
    def create_client(
        cls,
        hostname: str,
        qido_prefix: Optional[str] = None,
        wado_prefix: Optional[str] = None,
        auth: Optional[Union[AuthBase, Session]] = None,
    ):
        """Create a WsiDicomWebClient.

        Parameters
        ----------
        hostname: str
            The URL of the DICOMweb server
        qido_prefix: Optional[str]
            If needed by the server, provide the prefix for QIDO services
        wado_prefix: Optional[str]
            If needed by the server, provide the prefix for WADO services
        auth: Optional[Union[AuthBase, Session]]
            If needed by the server, provide authentication credentials.
            This may be provided by either passing an object that
            inherits from requests.auth.AuthBase, or by passing a
            requests.Session object.
        """
        if isinstance(auth, AuthBase):
            session = create_session_from_auth(auth)
        else:
            session = auth

        client = DICOMwebClient(
            hostname,
            qido_url_prefix=qido_prefix,
            wado_url_prefix=wado_prefix,
            session=session,
        )

        return cls(client)

    def get_wsi_instances(
        self, study_uid: UID, series_uids: Iterable[UID]
    ) -> Iterator[Tuple[UID, UID, Optional[Set[UID]]]]:
        """
        Get instance uids for WSI instances in a study.

        Parameters
        ----------
        study_uid: UID
            Study UID of the study.
        series_uids: Iterable[UID]
            Series UIDs in the study.

        Returns
        -------
        Iterator[Tuple[UID, UID, Optional[Set[UID]]]]
            Iterator of series and instance uid and optionally available transfer syntax
             uids for WSI instances in the study and series.
        """
        return self._get_intances(
            study_uid, series_uids, WSI_SOP_CLASS_UID, WSI_MODALITY
        )

    def get_annotation_instances(
        self, study_uid: UID, series_uids: Iterable[UID]
    ) -> Iterator[Tuple[UID, UID, Optional[Set[UID]]]]:
        """
        Get instance uids of Annotation instances in a study.

        Parameters
        ----------
        study_uid: UID
            Study UID of the study.
        series_uids: Iterable[UID]
            Series UIDs in the study.

        Returns
        -------
        Iterator[Tuple[UID, UID, Optional[Set[UID]]]]
            Iterator of series and instance uid and optionally available transfer syntax
            uids for Annotation instances in the study and series.
        """
        return self._get_intances(
            study_uid, series_uids, ANN_SOP_CLASS_UID, ANNOTATION_MODALITY
        )

    def get_instance(
        self, study_uid: UID, series_uid: UID, instance_uid: UID
    ) -> Dataset:
        """
        Get instance metadata.

        Parameters
        ----------
        study_uid: UID
            Study UID of the instance.
        series_uid: UID
            Series UID of the instance.
        instance_uid: UID
            Instance UID of the instance.

        Returns
        -------
        Dataset
            Instance metadata.
        """
        instance = self._client.retrieve_instance_metadata(
            study_uid, series_uid, instance_uid
        )
        return Dataset.from_json(instance)

    def get_frames(
        self,
        study_uid: UID,
        series_uid: UID,
        instance_uid: UID,
        frame_indices: List[int],
        transfer_syntax: UID,
    ) -> Iterator[bytes]:
        """
        Get frames from an instance.

        Parameters
        ----------
        study_uid: UID
            Study UID of the instance.
        series_uid: UID
            Series UID of the instance.
        instance_uid: UID
            Instance UID of the instance.
        frame_indices: List[int]
            List of frame indices to get. Note frames are indexed starting from 1.
        transfer_syntax: UID
            Transfer syntax of the to get frames.

        Returns
        -------
        Iterator[bytes]
            Iterator of frames.
        """
        return self._client.iter_instance_frames(
            study_uid,
            series_uid,
            instance_uid,
            frame_numbers=frame_indices,
            media_types=(self._transfer_syntax_to_media_type(transfer_syntax),),
        )

    def is_transfer_syntax_supported(
        self, study_uid: UID, series_uid: UID, instance_uid: UID, transfer_syntax: UID
    ) -> bool:
        """Check if transfer syntax is supported for retrieving frames.

        Parameters
        ----------
        study_uid: UID
            Study UID of the instance.
        series_uid: UID
            Series UID of the instance.
        instance_uid: UID
            Instance UID of the instance.
        transfer_syntax: UID
            Transfer syntax to check.

        Returns
        -------
        bool
            True if transfer syntax is supported, False otherwise.
        """
        try:
            self.get_frames(study_uid, series_uid, instance_uid, [1], transfer_syntax)
        except HTTPError as exception:
            if (
                exception.response is not None
                and exception.response.status_code == HTTPStatus.NOT_ACCEPTABLE
            ):
                logging.debug(
                    f"Transfer syntax {transfer_syntax} not supported "
                    f"for {instance_uid}."
                )
                return False
            raise exception
        logging.debug(
            f"Transfer syntax {transfer_syntax} supported for {instance_uid}."
        )
        return True

    def _get_intances(
        self,
        study_uid: UID,
        series_uids: Iterable[UID],
        sop_class_uid: UID,
        modality: str,
    ) -> Iterator[Tuple[UID, UID, Optional[Set[UID]]]]:
        """Get series, instance, and optionally available transfer syntax uids for
        instances of SOP class in study.

        Parameters
        ----------
        study_uid: UID
            Study UID of the study.
        series_uids: Iterable[UID]
            Series UIDs in the study.
        sop_class_uid: UID
            SOP Class UID of the instances.

        Returns
        -------
        Iterator[Tuple[UID, UID, Optional[Set[UID]]]]
            Iterator of series and instance uid and optionally available transfer syntax
            uids for instances in the study and series.
        """
        if isinstance(self._client, DICOMfileClient):
            # DICOMfileClient does not support searching for instances by
            # series instance uid as search filter
            return (
                self._get_uids_from_response(instance, series_uid)
                for series_uid in series_uids
                for instance in self._client.search_for_instances(
                    study_uid,
                    series_uid,
                    search_filters={SOP_CLASS_UID: sop_class_uid, MODALITY: modality},
                )
            )
        return (
            self._get_uids_from_response(instance)
            for instance in self._search_for_instances(
                study_uid,
                fields=["AvailableTransferSyntaxUID"],
                search_filters={
                    SOP_CLASS_UID: sop_class_uid,
                    SERIES_INSTANCE_UID: series_uids,
                    MODALITY: modality,
                },
            )
        )

    def _search_for_instances(
        self, study_uid: UID, fields: List[str], search_filters: Dict[str, Any]
    ) -> Iterator[Dict[str, Dict[Any, Any]]]:
        """Search for instances in study.

        Catches known errors in server DICOMweb implementation and tries to fix them.

        Parameters
        ----------
        study_uid: UID
            Study UID of the study.
        fields: List[str]
            Fields to include in the response.
        search_filters: Dict[str, Any]
            Search filters to use.

        Returns
        -------
        Iterator[Dict[str, Dict[Any, Any]]]
            Iterator of instance metadata.
        """
        # Errors that can be fixed by removing the offending filter and filtering
        # the results.
        known_search_filter_errors = {
            HTTPStatus.BAD_REQUEST: {
                "SOPClassUID is not a supported instance": SOP_CLASS_UID
            }
        }
        # Errors that can be fixed by removing the offending field. Should only
        # be used if the offending field is not required.
        known_field_errors = {
            HTTPStatus.BAD_REQUEST: {
                "unknown/unsupported QIDO attribute: AvailableTransferSyntaxUID": "AvailableTransferSyntaxUID"
            }
        }
        try:
            return iter(
                self._client.search_for_instances(
                    study_uid,
                    fields=fields,
                    search_filters=search_filters,
                )
            )
        except HTTPError as exception:
            status_code = HTTPStatus(exception.response.status_code)
            error_message = exception.response.text
            logging.debug(
                f"Got error code: {status_code} message: {error_message} "
                "when searching for instances."
            )
            # If search filter error remove offending filter and filter the results.
            try:
                filter_key = next(
                    filter_key
                    for error_key, filter_key in known_search_filter_errors[
                        status_code
                    ].items()
                    if status_code in known_search_filter_errors
                    if error_key in error_message
                )
                logging.debug(f"Removing filter {filter_key} from search filters.")
                filter_value = search_filters.pop(filter_key)
                instances = self._search_for_instances(
                    study_uid,
                    fields=fields,
                    search_filters=search_filters,
                )
                # Filter out instances with the removed search filter.
                return (
                    instance
                    for instance in instances
                    if instance[filter_key]["Value"][0] == filter_value
                )
            except StopIteration:
                pass

            # If a field error remove offending field.
            try:
                field_key = next(
                    field_key
                    for error_key, field_key in known_field_errors[status_code].items()
                    if status_code in known_search_filter_errors
                    if error_key in error_message
                )
                logging.debug(f"Removing field {field_key} from fields.")
                fields.remove(field_key)
                return self._search_for_instances(
                    study_uid,
                    fields=fields,
                    search_filters=search_filters,
                )
            except StopIteration:
                pass
            # Not a known error. Propagate the exception.
            raise

    @staticmethod
    def _get_uids_from_response(
        response: Dict[str, Dict[Any, Any]], series_uid: Optional[UID] = None
    ) -> Tuple[UID, UID, Optional[Set[UID]]]:
        """Get series, instance, and optionally transfer syntax uids from response.

        Parameters
        ----------
        response: Dict[str, Dict[Any, Any]]
            Response from server for an instance.

        Returns
        -------
        Tuple[UID, UID, Optional[Set[UID]]]
            Series and instance uid and optionally available transfer syntax
            uids for instances in response.
        """
        available_transfer_syntaxes = response.get(
            AVAILABLE_SOP_TRANSFER_SYNTAX_UID, None
        )
        return (
            (
                series_uid
                if series_uid is not None
                else UID(response[SERIES_INSTANCE_UID]["Value"][0])
            ),
            UID(response[SOP_INSTANCE_UID]["Value"][0]),
            (
                set(available_transfer_syntaxes["Value"])
                if available_transfer_syntaxes
                else None
            ),
        )

    @staticmethod
    def _get_sop_class_uid_from_response(response: Dict[str, Dict[Any, Any]]) -> UID:
        """Get SOP Class UID from response.

        Parameters
        ----------
        response: Dict[str, Dict[Any, Any]]
            Response from server.

        Returns
        -------
        UID
            SOP Class UID from response.
        """
        return UID(response[SOP_CLASS_UID]["Value"][0])

    @staticmethod
    def _transfer_syntax_to_media_type(transfer_syntax: UID) -> Tuple[str, str]:
        """Convert transfer syntax to media type.

        Parameters
        ----------
        transfer_syntax: UID
            Transfer syntax to convert.

        Returns
        -------
        Tuple[str, str]
            Media type and transfer syntax.
        """
        try:
            return determine_media_type(transfer_syntax), transfer_syntax
        except NotImplementedError as exception:
            raise ValueError(
                f"Could not determine media type for transfer syntax {transfer_syntax}"
            ) from exception



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Image model."""

import datetime
from dataclasses import dataclass, replace
from enum import Enum
from typing import Optional, Sequence, TypeVar

from wsidicom.codec import LossyCompressionIsoStandard
from wsidicom.geometry import Orientation, PointMm, RegionMm, SizeMm


class FocusMethod(Enum):
    """Focus methods."""

    AUTO = "auto"
    MANUAL = "manual"


@dataclass(frozen=True)
class ExtendedDepthOfField:
    """Extended depth of field.

    Parameters
    ----------
    number_of_focal_planes : int
        The number of focal planes used for extended depth of field.
    distance_between_focal_planes : float
        The distance between the focal planes used for extended depth of field.
    """

    number_of_focal_planes: int
    distance_between_focal_planes: float


GeometryType = TypeVar("GeometryType", PointMm, RegionMm)


@dataclass(frozen=True)
class ImageCoordinateSystem:
    """Image coordinate system.

    Parameters
    ----------
    origin : PointMm
        The position of the top left pixel in the slide coorindate system.
    rotation : float
        The rotation of the image in degrees in the slide coordinate system.
    """

    origin: PointMm
    rotation: float

    @property
    def orientation(self) -> Orientation:
        return Orientation.from_rotation(self.rotation)

    def image_to_slide(self, image: GeometryType) -> GeometryType:
        """Translate a geometry (PointMm or RegionMm) from image coordinate system
        (origin at image top-left corner) to slide coordinate system (origin at
        slide corner).

        Parameters
        ----------
        image : GeometryType
            The geometry to translate.

        Returns
        -------
        GeometryType
            The translate geometry.
        """
        if isinstance(image, PointMm):
            offset = self.orientation.apply_transform(image)
            return self.origin + offset
        start = self.image_to_slide(image.start)
        end = self.image_to_slide(image.end)
        return RegionMm(start, SizeMm(end.x - start.x, end.y - start.y))

    def slide_to_image(self, slide: GeometryType) -> GeometryType:
        """Translate a geometry (PointMm or RegionMm) from slide coordinate system
        (origin at slide corner) to image coordinate system (origin at image top-left
        corner).

        Parameters
        ----------
        slide : GeometryType
            The geometry to translate.

        Returns
        -------
        GeometryType
            The translate geometry.
        """
        if isinstance(slide, PointMm):
            offset = slide - self.origin
            return self.orientation.apply_reverse_transform(offset)
        start = self.slide_to_image(slide.start)
        end = self.slide_to_image(slide.end)
        start, end = (
            PointMm(min(start.x, end.x), min(start.y, end.y)),
            PointMm(max(start.x, end.x), max(start.y, end.y)),
        )
        size = SizeMm(end.x - start.x, end.y - start.y)
        return RegionMm(start, size)

    def to_other_coordinate_system(
        self, other: "ImageCoordinateSystem", image: GeometryType
    ) -> GeometryType:
        """Translate a geometry (PointMm or RegionMm) from this image coordinate system
        to another image coordinate system.

        Parameters
        ----------
        other : ImageCoordinateSystem
            The target image coordinate system.
        image : GeometryType
            The geometry to translate.

        Returns
        -------
        GeometryType
            The translate geometry.
        """
        slide = self.image_to_slide(image)
        return other.slide_to_image(slide)


@dataclass(frozen=True)
class LossyCompression:
    """Lossy compression.

    Parameters
    ----------
    method : LossyCompressionIsoStandard
        The method used for lossy compression.
    ratio : float
        The compression ratio.
    """

    method: LossyCompressionIsoStandard
    ratio: float


@dataclass(frozen=True)
class Image:
    """
    Image metadata.

    Corresponds to the `Required, Empty if Unknown` attributes in the Slide Label
    module:
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.8.12.8.html

    Parameters
    ----------
    acquisition_datetime : Optional[datetime.datetime] = None
        The acquisition datetime of the image.
    focus_method : Optional[FocusMethod] = None
        The focus method used for imaging.
    extended_depth_of_field : Optional[ExtendedDepthOfField] = None
        Describes if extended depth of field has been used for imaging.
    image_coordinate_system : Optional[ImageCoordinateSystem] = None
        The image coordinate system in relation the slide frame of reference.
    pixel_spacing : Optional[SizeMm] = None
        The pixel spacing in mm per pixel.
    focal_plane_spacing : Optional[float] = None
        The spacing between focal planes if image with multiple focal planes.
    depth_of_field : Optional[float] = None
        The depth of field of the image.
    lossy_compressions : Optional[Sequence[LossyCompression]] = None
        The lossy compressions method that has been applied to the image data.
    """

    acquisition_datetime: Optional[datetime.datetime] = None
    focus_method: Optional[FocusMethod] = None
    extended_depth_of_field: Optional[ExtendedDepthOfField] = None
    image_coordinate_system: Optional[ImageCoordinateSystem] = None
    pixel_spacing: Optional[SizeMm] = None
    focal_plane_spacing: Optional[float] = None
    depth_of_field: Optional[float] = None
    lossy_compressions: Optional[Sequence[LossyCompression]] = None

    def remove_confidential(self) -> "Image":
        return replace(
            self,
            acquisition_datetime=None,
        )



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Models for specimen description."""

import datetime
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Any, Iterable, List, Optional, Sequence, Tuple, Union

from pydicom.sr.coding import Code
from pydicom.uid import UID, generate_uid

from wsidicom.conceptcode import (
    AnatomicPathologySpecimenTypesCode,
    ContainerTypeCode,
    SpecimenCollectionProcedureCode,
    SpecimenEmbeddingMediaCode,
    SpecimenFixativesCode,
    SpecimenPreparationStepsCode,
    SpecimenSamplingProcedureCode,
    SpecimenStainsCode,
    UnitCode,
)


class UniversalIssuerType(Enum):
    """The type of a universal issuer of an identifier."""

    DNS = "DNS"  # An Internet dotted name. Either in ASCII or as integers
    EUI64 = "EUI64"  # An IEEE Extended Unique Identifier
    ISO = "ISO"  # An International Standards Organization Object Identifier
    URI = "IRI"  # Uniform Resource Identifier
    UUID = "UUID"  # The DCE Universal Unique Identifier
    X400 = "X400"  # An X.400 MHS identifier
    X500 = "X500"  # An X.500 directory name


class IssuerOfIdentifier(metaclass=ABCMeta):
    """Metaclass for an issuer of an issuer of identifier."""

    @abstractmethod
    def to_hl7v2(self) -> str:
        """Return HL7v2 representation of issuer."""
        raise NotImplementedError()

    @classmethod
    def from_hl7v2(cls, value: str) -> "IssuerOfIdentifier":
        """Return IssuerOfIdentifier from HL7v2 representation."""
        local_identifier, *universal_identifier = value.split("^")
        if local_identifier == "":
            local_identifier = None
        if len(universal_identifier) == 2:
            universal_identifier, universal_issuer_type = universal_identifier
            return UniversalIssuerOfIdentifier(
                identifier=universal_identifier,
                issuer_type=UniversalIssuerType(universal_issuer_type),
                local_identifier=local_identifier,
            )
        if local_identifier is not None:
            return LocalIssuerOfIdentifier(identifier=value)
        raise ValueError(
            "Could not parse string to local issuer of identifier or "
            "universal issuer of identifier."
        )


@dataclass(frozen=True)
class LocalIssuerOfIdentifier(IssuerOfIdentifier):
    """A local issuer of an identifier.

    Parameters
    ----------
    identifier: str
        The identifier of the local issuer.
    """

    identifier: str

    def to_hl7v2(self) -> str:
        return self.identifier


@dataclass(frozen=True)
class UniversalIssuerOfIdentifier(IssuerOfIdentifier):
    """A universal issuer of an identifier.

    Parameters
    ----------
    identifier: str
        The identifier of the universal issuer.
    issuer_type: UniversalIssuerType
        The type of the universal issuer.
    local_identifier: Optional[str] = None
        Optional local identifier of the universal issuer.
    """

    identifier: str
    issuer_type: UniversalIssuerType
    local_identifier: Optional[str] = None

    def to_hl7v2(self) -> str:
        identifier = self.local_identifier or ""
        identifier += f"^{self.identifier}^{self.issuer_type.name}"
        return identifier


@dataclass(frozen=True)
class Measurement:
    """A measurement with a value and a unit.

    Parameters
    ----------
    value: float
        The value of the measurement.
    unit: UnitCode
        The unit of the measurement.
    """

    value: float
    unit: UnitCode


@dataclass(frozen=True)
class SamplingLocation:
    """The location of a sampling.

    Parameters
    ----------
    reference: Optional[str] = None
        Description of coordinate system and origin reference point used for the
        location.
    description: Optional[str] = None
        Description of the location.
    x: Optional[Measurement] = None
        The x-coordinate of the location.
    y: Optional[Measurement] = None
        The y-coordinate of the location.
    z: Optional[Measurement] = None
        The z-coordinate of the location.
    """

    reference: Optional[str] = None
    description: Optional[str] = None
    x: Optional[Measurement] = None
    y: Optional[Measurement] = None
    z: Optional[Measurement] = None


@dataclass(frozen=True)
class SampleLocalization(SamplingLocation):
    """The location of a sample on a slide.

    Parameters
    ----------
    reference: Optional[str] = None
        Description of coordinate system and origin reference point used for the
        location.
    description: Optional[str] = None
        Description of the location.
    x: Optional[Measurement] = None
        The x-coordinate of the location.
    y: Optional[Measurement] = None
        The y-coordinate of the location.
    z: Optional[Measurement] = None
        The z-coordinate of the location.
    visual_marking: Optional[str] = None
        Description of visual marking of the sample, for example ink or shape.
    """

    visual_marking: Optional[str] = None


@dataclass(frozen=True)
class SpecimenIdentifier:
    """A specimen identifier including an optional issuer.

    Parameters
    ----------
    value: str
        The value of the identifier.
    issuer: Optional[IssuerOfIdentifier] = None
        Optional issuer of the identifier.
    """

    value: str
    issuer: Optional[IssuerOfIdentifier] = None

    def __eq__(self, other: Any):
        """Determine if other specimen identifiers is equal to this."""
        if isinstance(other, str):
            return self.value == other and self.issuer is None
        if isinstance(other, SpecimenIdentifier):
            return self.value == other.value and self.issuer == other.issuer
        return False

    def to_string_identifier_and_issuer(self) -> Tuple[str, Optional[str]]:
        """Format into string identifier and IssuerOfIdentifier."""
        if self.issuer is None:
            return self.value, None
        return self.value, self.issuer.to_hl7v2()

    @classmethod
    def get_string_identifier_and_issuer(
        cls, identifier: Union[str, "SpecimenIdentifier"]
    ) -> Tuple[str, Optional[str]]:
        """Return string identifier and optional issuer of identifier object."""
        if isinstance(identifier, str):
            return identifier, None
        return identifier.to_string_identifier_and_issuer()

    @classmethod
    def get_identifier_and_issuer(
        cls, identifier: Union[str, "SpecimenIdentifier"]
    ) -> Tuple[str, Optional[IssuerOfIdentifier]]:
        """Return string identifier and optional issuer of identifier object."""
        if isinstance(identifier, str):
            return identifier, None
        return identifier.value, identifier.issuer

    def simplify(self) -> Union[str, "SpecimenIdentifier"]:
        """Return a simplified version of the identifier."""
        if self.issuer is None:
            return self.value
        return self

    def relaxed_matching(self, other: "SpecimenIdentifier") -> bool:
        """Determine if other specimen identifiers is equal to this, but does not
        require issuer to be the same if one of the issuers is None."""
        if self == other:
            return True
        if self.value != other.value:
            return False
        return self.issuer is None or other.issuer is None


class PreparationStep(metaclass=ABCMeta):
    """
    Metaclass for a preparation step for a specimen.

    A preparation step is an action performed on a specimen.
    """


class BaseSampling(PreparationStep, metaclass=ABCMeta):
    """Either a `Sampling` or a `UnknownSampling`."""

    specimen: "BaseSpecimen"
    sampling_constraints: Optional[Sequence["BaseSampling"]]


@dataclass(frozen=True)
class Sampling(BaseSampling):
    """
    The sampling of a specimen into samples that can be used to create new specimens.

    Parameters
    ----------
    specimen: BaseSpecimen
        The specimen that was sampled.
    method: SpecimenSamplingProcedureCode
        Method used for sampling. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8110.html
        or `SpecimenSamplingProcedureCode.meanings` for allowed sampling methods.
    specimen_type: AnatomicPathologySpecimenTypesCode
        The type of the specimen. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8103.html
         or `AnatomicPathologySpecimenTypesCode.meanings` for allowed specimen types.
    sampling_constraints: Optional[Sequence[BaseSampling]] = None
        Optional constraints on the sampling tree. The constraints can be used to
        define a single branch from the sample to a end specimen in the sampling tree.
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the sampling was performed.
    description: Optional[str] = None
        Optional description of the sampling.
    location: Optional[SamplingLocation] = None
        Optional location of the sampling in the parent specimen.
    """

    specimen: "BaseSpecimen"
    method: SpecimenSamplingProcedureCode
    specimen_type: AnatomicPathologySpecimenTypesCode = field(repr=False)
    sampling_constraints: Optional[Sequence[BaseSampling]] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None
    location: Optional[SamplingLocation] = None


@dataclass(frozen=True)
class UnknownSampling(BaseSampling):
    """
    Represent an unknown relation between a parent specimen and a child sample.

    The relation between the parent and child can be direct (e.g. from block to slide)
    or several steps in between can be missing (e.g. from specimen to slide). If
    possible is preferred to use the `Sampling` class instead.

    Parameters
    ----------
    specimen: BaseSpecimen
        The specimen that was sampled.
    sampling_constraints: Optional[Sequence[BaseSampling]] = None
        Optional constraints on the sampling tree. The constraints should be used to
        define a single branch for the sample in the sampling tree.
    """

    specimen: "BaseSpecimen"
    sampling_constraints: Optional[Sequence[BaseSampling]] = None


@dataclass(frozen=True)
class Collection(PreparationStep):
    """
    The collection of a specimen from a body.

    Parameters
    ----------
    method: SpecimenCollectionProcedureCode
        Method used for collection. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8109.html
        or `SpecimenCollectionProcedureCode.meanings` for allowed collection methods.
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the collection was performed.
    description: Optional[str] = None
        Optional description of the collection.
    """

    method: SpecimenCollectionProcedureCode
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


@dataclass(frozen=True)
class Processing(PreparationStep):
    """
    Other processing steps, such as heating or clearing, made on a specimen.

    Parameters
    ----------
    method: SpecimenPreparationStepsCode
        Method used for processing. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8113.html
        or `SpecimenPreparationStepsCode.meanings` for allowed processing methods.
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the processing was performed.
    description: Optional[str] = None
        Optional description of the processing.
    """

    method: Optional[SpecimenPreparationStepsCode] = None
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


@dataclass(frozen=True)
class Embedding(PreparationStep):
    """
    Embedding of a specimen in an medium.

    Parameters
    ----------
    medium: SpecimenEmbeddingMediaCode
        The medium used for embedding. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8115.html
        or `SpecimenEmbeddingMediaCode.meanings` for allowed mediums.
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the embedding was performed.
    description: Optional[str] = None
        Optional description of the embedding.
    """

    medium: SpecimenEmbeddingMediaCode
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


@dataclass(frozen=True)
class Fixation(PreparationStep):
    """
    Fixation of a specimen using a fixative.

    Parameters
    ----------
    fixative: SpecimenFixativesCode
        The fixative used for fixation. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8114.html
        or `SpecimenFixativesCode.meanings` for allowed fixatives.
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the fixation was performed.
    description: Optional[str] = None
        Optional description of the fixation.
    """

    fixative: SpecimenFixativesCode
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


@dataclass(frozen=True)
class Staining(PreparationStep):
    """
    Staining of a specimen using staining substances.

    Parameters
    ----------
    substances: Union[str, Sequence[SpecimenStainsCode]]
        The substances used for staining. Either a single string or a list of codes. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8112.html
        or SpecimenStainsCode.meanings` for allowed stain codes.
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the staining was performed.
    description: Optional[str] = None
        Optional description of the staining.
    """

    substances: Union[str, Sequence[SpecimenStainsCode]]
    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


@dataclass(frozen=True)
class Receiving(PreparationStep):
    """
    Reception of a specimen.

    Parameters
    ----------
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the reception was performed.
    description: Optional[str] = None
        Optional description of the reception.
    """

    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


@dataclass(frozen=True)
class Storage(PreparationStep):
    """
    Storage of a specimen.

    Parameters
    ----------
    date_time: Optional[datetime.datetime] = None
        Optional date and time when the storage was performed.
    description: Optional[str] = None
        Optional description of the storage.
    """

    date_time: Optional[datetime.datetime] = None
    description: Optional[str] = None


class BaseSpecimen(metaclass=ABCMeta):
    """Metaclass for a specimen."""

    steps: List[PreparationStep]
    identifier: Union[str, SpecimenIdentifier]
    type: Optional[AnatomicPathologySpecimenTypesCode]
    container: Optional[ContainerTypeCode]

    @property
    def samplings(self) -> List[Sampling]:
        """Return list of samplings done on the specimen."""
        return [step for step in self.steps if isinstance(step, Sampling)]

    def add(self, step: PreparationStep) -> None:
        """Add a preparation step to the sequence of steps for the specimen."""
        if not isinstance(step, BaseSampling):
            if any(
                isinstance(present_step, BaseSampling) for present_step in self.steps
            ):
                raise ValueError(
                    "Only additional samplings can be added to a specimen that has "
                    "been sampled."
                )
        self.steps.append(step)


class SampledSpecimen(BaseSpecimen, metaclass=ABCMeta):
    """Metaclass for a specimen thas has been sampled from one or more specimens."""

    sampled_from: Optional[Union[BaseSampling, List[BaseSampling]]]

    @property
    def sampled_from_list(self) -> List[BaseSampling]:
        if self.sampled_from is None:
            return []
        if isinstance(self.sampled_from, BaseSampling):
            return [self.sampled_from]
        return self.sampled_from

    def add(self, step: PreparationStep) -> None:
        if isinstance(step, Collection):
            raise ValueError(
                "A collection step can only be added to specimens of type `Specimen`"
            )
        super().add(step)

    def _check_sampling_constraints_in_sampling_tree(
        self, constraints: Optional[Sequence[BaseSampling]]
    ) -> None:
        """Check that the sampling tree for the specimen contains the constraints."""
        if constraints is None:
            return

        def recursive_search(sampling: BaseSampling, specimen: BaseSpecimen) -> bool:
            """Recursively search for sampling in samplings for specimen."""
            if sampling in specimen.samplings:
                return True
            if isinstance(specimen, SampledSpecimen):
                return any(
                    recursive_search(sampling, sampling.specimen)
                    for sampling in specimen.sampled_from_list
                )
            return False

        for constraint in constraints:
            if not recursive_search(constraint, self):
                raise ValueError(
                    "Could not create sampling as specimen was not sampled "
                    f"from {constraint}"
                )

    def is_sampling_constraint_is_unambiguous(
        self,
        constraints: Iterable[Union[str, SpecimenIdentifier]],
    ):
        """Check that the sampling tree for the specimen is is not branching when
        following the constraints."""
        if len(self.sampled_from_list) == 0:
            # No sampling tree, is unambiguous
            return True
        if len(self.sampled_from_list) > 1:
            # Check that at one and only one of the samplings are in the constraints
            samplings_in_constraints = (
                sampling
                for sampling in self.sampled_from_list
                if sampling.specimen.identifier in constraints
            )
            first_sampling_in_constraints = next(samplings_in_constraints, None)
            if first_sampling_in_constraints is None:
                return False
            if next(samplings_in_constraints, None) is not None:
                return False
            sampling_to_check = first_sampling_in_constraints
        else:
            # Only one sampling to check
            sampling_to_check = self.sampled_from_list[0]

        if not isinstance(sampling_to_check.specimen, SampledSpecimen):
            # Sampling tree ends non-sampled specimen, is unambiguous
            return True

        # Update the constrain and check the parent specimen
        sub_constraint = set(constraints)
        if sampling_to_check.sampling_constraints is not None:
            sub_constraint.update(
                constraint.specimen.identifier
                for constraint in sampling_to_check.sampling_constraints
            )
        return sampling_to_check.specimen.is_sampling_constraint_is_unambiguous(
            sub_constraint
        )


@dataclass(frozen=True)
class Specimen(BaseSpecimen):
    """A specimen that has been extracted/taken from a patient in some way. Does not
    need to represent the actual first specimen in the collection tree, but should
    represent the first known (i.e. that we have metadata for) specimen in the
    collection tree.

    Parameters
    ----------
    identifier: Union[str, SpecimenIdentifier]
        The identifier of the specimen.
    extraction_step: Optional[Collection] = None
        Optional collection step for the specimen.
    type: Optional[AnatomicPathologySpecimenTypesCode] = None
        Optional type of the specimen. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8103.html
        or `AnatomicPathologySpecimenTypesCode.meanings` for allowed specimen types.
    steps: Sequence[PreparationStep] = []
        Optional sequence of preparation steps for the specimen.
    container: Optional[ContainerTypeCode] = None
        Optional container type of the specimen. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8101.html
        or `ContainerTypeCode.meanings` for allowed container types.
    """

    identifier: Union[str, SpecimenIdentifier]
    extraction_step: Optional[Collection] = None
    type: Optional[AnatomicPathologySpecimenTypesCode] = None
    steps: List[PreparationStep] = field(default_factory=list)
    container: Optional[ContainerTypeCode] = None

    def __post_init__(self):
        if self.extraction_step is not None:
            if len(self.steps) == 0 or (
                len(self.steps) > 0 and self.steps[0] != self.extraction_step
            ):
                self.steps.insert(
                    0,
                    self.extraction_step,
                )

    def add(self, step: PreparationStep) -> None:
        if isinstance(step, Collection) and len(self.steps) != 0:
            raise ValueError("A Collection-step must be the first step.")
        super().add(step)

    def sample(
        self,
        method: Optional[SpecimenSamplingProcedureCode] = None,
        date_time: Optional[datetime.datetime] = None,
        description: Optional[str] = None,
        location: Optional[SamplingLocation] = None,
    ) -> BaseSampling:
        """Create a sampling from the specimen that can be used to create a new
        sample."""
        if method is None or self.type is None:
            sampling = UnknownSampling(
                specimen=self,
                sampling_constraints=None,
            )
        else:
            sampling = Sampling(
                specimen=self,
                method=method,
                specimen_type=self.type,
                sampling_constraints=None,
                date_time=date_time,
                description=description,
                location=location,
            )
        self.add(sampling)
        return sampling


@dataclass(frozen=True)
class Sample(SampledSpecimen):
    """A specimen that has been sampled from one or more other specimens.

    Parameters
    ----------
    identifier: Union[str, SpecimenIdentifier]
        The identifier of the sample.
    sampled_from: Sequence[BaseSampling]
        Sequence of samplings that the sample was made from.
    type: Optional[AnatomicPathologySpecimenTypesCode] = None
        Optional type of the sample. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8103.html
        or `AnatomicPathologySpecimenTypesCode.meanings` for allowed specimen types.
    steps: Sequence[PreparationStep] = []
        Optional sequence of preparation steps for the sample.
    container: Optional[ContainerTypeCode] = None
        Optional container type of the sample. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8101.html
        or `ContainerTypeCode.meanings` for allowed container types.
    """

    identifier: Union[str, SpecimenIdentifier]
    sampled_from: Sequence[BaseSampling]
    type: Optional[AnatomicPathologySpecimenTypesCode] = None
    steps: Sequence[PreparationStep] = field(default_factory=list)
    container: Optional[ContainerTypeCode] = None

    def sample(
        self,
        method: Optional[SpecimenSamplingProcedureCode] = None,
        date_time: Optional[datetime.datetime] = None,
        description: Optional[str] = None,
        sampling_constraints: Optional[Sequence[BaseSampling]] = None,
        location: Optional[SamplingLocation] = None,
    ) -> BaseSampling:
        """Create a sampling from the specimen that can be used to create a new
        sample."""
        self._check_sampling_constraints_in_sampling_tree(sampling_constraints)
        if sampling_constraints is not None:
            for sampling_constraint in sampling_constraints:
                assert isinstance(sampling_constraint, Sampling)

        if method is None or self.type is None:
            sampling = UnknownSampling(
                specimen=self,
                sampling_constraints=sampling_constraints,
            )
        else:
            sampling = Sampling(
                specimen=self,
                specimen_type=self.type,
                method=method,
                sampling_constraints=sampling_constraints,
                date_time=date_time,
                description=description,
                location=location,
            )
        self.add(sampling)
        return sampling


@dataclass(frozen=True)
class SlideSample(SampledSpecimen):
    """A sample that has been placed on a slide.

    Parameters
    ----------
    identifier: Union[str, SpecimenIdentifier]
        The identifier of the slide sample.
    anatomical_sites: Optional[Sequence[Code]] = None
        Optional primary anatomic structures of interest in the slide sample.
    sampled_from: Optional[BaseSampling] = None
        Optional sampling that the slide sample was made from.
    uid: Optional[UID] = None
        Optional UID of the slide sample.
    localization: Optional[SampleLocalization] = None
        Optional localization of the slide sample on the slide.
    steps: Sequence[PreparationStep] = []
        Optional sequence of preparation steps for the slide sample.
    short_description: Optional[str] = None
        Optional short description of the slide sample. Should max be 64 characters.
    detailed_description: Optional[str] = None
        Optional detailed description of the slide sample.
    """

    identifier: Union[str, SpecimenIdentifier]
    anatomical_sites: Optional[Sequence[Code]] = None
    sampled_from: Optional[BaseSampling] = None
    uid: Optional[UID] = None
    localization: Optional[SampleLocalization] = None
    steps: List[PreparationStep] = field(default_factory=list)
    short_description: Optional[str] = None
    detailed_description: Optional[str] = None
    container: ContainerTypeCode = field(
        init=False, default=ContainerTypeCode("Microscope slide")
    )
    type: AnatomicPathologySpecimenTypesCode = field(
        init=False, default=AnatomicPathologySpecimenTypesCode("Slide")
    )

    def __post_init__(self):
        self._check_slide_sample_sampling_constraints()

    def _check_slide_sample_sampling_constraints(self):
        """Check that the sampling tree for the slide sample is not branching."""
        if self.sampled_from is None or not isinstance(
            self.sampled_from.specimen, SampledSpecimen
        ):
            # No sampling constraints possible
            return
        if not self.sampled_from.specimen.is_sampling_constraint_is_unambiguous(
            [
                constraint.specimen.identifier
                for constraint in self.sampled_from.sampling_constraints
            ]
            if self.sampled_from.sampling_constraints is not None
            else []
        ):
            raise ValueError("Sampling constraints for slide sample are ambiguous.")

    @cached_property
    def default_uid(self) -> UID:
        """Uid used if not set."""
        if self.uid is not None:
            return self.uid
        return generate_uid()



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Frame index for empty BOT, parsing the positions from the pixel data."""

from typing import List, Tuple

from pydicom.tag import ItemTag

from wsidicom.errors import WsiDicomFileError
from wsidicom.file.io.frame_index.bot import EmptyBotException
from wsidicom.file.io.frame_index.encapsulated_pixel_data import EncapsulatedPixelData
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType


class EmptyBot(EncapsulatedPixelData):
    """Frame index parsed from reading the sequence of pixel data delimeters."""

    @property
    def offset_table_type(self) -> OffsetTableType:
        return OffsetTableType.EMPTY

    def _get_index(self) -> List[Tuple[int, int]]:
        """Get frame positions and length from sequence of frames that ends with a tag
        not equal to ItemTag.

        Each frame contains:
        item tag (4 bytes)
        item length (4 bytes)
        item data (item length)
        The position of item data and the item length is stored.

        Returns
        -------
        list[tuple[int, int]]
            A list with frame positions and frame lengths
        """
        self._file.seek(self._pixels_start)
        TAG_BYTES = 4
        LENGTH_BYTES = 4
        positions: List[Tuple[int, int]] = []
        frame_position = self._file.tell()
        # Read items until sequence delimiter
        while self._file.read_le_tag() == ItemTag:
            # Read item length
            length: int = self._file.read_leUL()
            if length == 0 or length % 2:
                raise WsiDicomFileError(str(self._file), "Invalid frame length")
            positions.append((frame_position + TAG_BYTES + LENGTH_BYTES, length))
            # Jump to end of frame
            self._file.seek(length, 1)
            frame_position = self._file.tell()
        self._file.read_sequence_delimiter()
        return positions

    def _get_pixels_start(self) -> int:
        self._validate_pixel_data_start()
        bot_length = self._read_bot_length()
        if bot_length is not None:
            raise EmptyBotException()
        return self._file.tell()



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Study model."""

import datetime
from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from pydicom.uid import UID, generate_uid


@dataclass(frozen=True)
class Study:
    """
    Study metadata.

    Corresponds to the `Required` and `Required, Empty if Unknown` attributes in the
    General Study Module:
    https://dicom.nema.org/medical/Dicom/current/output/chtml/part03/sect_C.7.2.html

    Parameters
    ----------
    uid : Optional[UID] = None
        The study instance UID.
    identifier : Optional[str] = None
        The study identifier (study ID).
    date : Optional[datetime.date] = None
        The date the study was performed.
    time : Optional[datetime.time] = None
        The time the study was performed.
    accession_number : Optional[str] = None
        The accession number of the study.
    referring_physician_name : Optional[str] = None
        The name of the referring physician.
    """

    uid: Optional[UID] = None
    identifier: Optional[str] = None
    date: Optional[datetime.date] = None
    time: Optional[datetime.time] = None
    accession_number: Optional[str] = None
    referring_physician_name: Optional[str] = None

    @cached_property
    def default_uid(self) -> UID:
        if self.uid is not None:
            return self.uid
        return generate_uid()



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Json schema for Label model."""

from marshmallow import fields

from wsidicom.metadata.label import Label
from wsidicom.metadata.schema.common import LoadingSchema


class LabelJsonSchema(LoadingSchema[Label]):
    text = fields.String(allow_none=True)
    barcode = fields.String(allow_none=True)
    label_in_volume_image = fields.Boolean(load_default=False, allow_none=True)
    label_in_overview_image = fields.Boolean(load_default=False, allow_none=True)
    label_is_phi = fields.Boolean(load_default=True, allow_none=True)

    @property
    def load_type(self):
        return Label



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""DICOM fields for attribute serialization."""

import datetime
import logging
import math
from abc import abstractmethod
from typing import (
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from marshmallow import Schema, fields, post_dump, post_load, pre_dump, pre_load
from marshmallow.utils import missing
from pydicom import DataElement, Dataset
from pydicom import config as pydicom_config
from pydicom.multival import MultiValue
from pydicom.sr.coding import Code
from pydicom.uid import UID
from pydicom.valuerep import (
    DA,
    DT,
    MAX_VALUE_LEN,
    STR_VR,
    TM,
    VR,
    DSfloat,
    PersonName,
    validate_vr_length,
)

from wsidicom import config
from wsidicom.conceptcode import ConceptCode, UnitCode
from wsidicom.geometry import Orientation, PointMm, SizeMm
from wsidicom.metadata.sample import (
    IssuerOfIdentifier,
    LocalIssuerOfIdentifier,
    Measurement,
    UniversalIssuerOfIdentifier,
    UniversalIssuerType,
)


class StringLikeDicomField(fields.Field):
    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        """By default pydicom returns empty string for empty string-like elements."""
        deserialized = super()._deserialize(value, attr, data, **kwargs)
        if deserialized == "":
            return None
        return deserialized


class StringDicomField(StringLikeDicomField):
    def __init__(self, value_representation: VR, **kwargs):
        if value_representation not in STR_VR:
            raise ValueError(
                f"Value representation {value_representation} is not a string."
            )
        self._value_representation = value_representation
        super().__init__(**kwargs)

    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if value is None:
            return None
        valid, error = validate_vr_length(self._value_representation, value)
        if (
            not valid
            and pydicom_config.settings.writing_validation_mode == pydicom_config.RAISE
            and config.settings.truncate_long_dicom_strings_on_validation_error
        ):
            maximum_allowed_legth = MAX_VALUE_LEN[self._value_representation]
            logging.warning(
                f"Truncating long DICOM string {value} of value representation "
                f"{self._value_representation} with maximum allowed length "
                f"{maximum_allowed_legth} to {value[: maximum_allowed_legth]}."
            )
            value = value[:maximum_allowed_legth]
        return super()._serialize(value, attr, obj, **kwargs)


class EnumDicomField(fields.Enum):
    def _deserialize(self, value, attr, data, **kwargs):
        if value == "":
            return None
        return super()._deserialize(value, attr, data, **kwargs)


class DateTimeDicomField(StringLikeDicomField):
    def _serialize(
        self,
        value: Optional[datetime.datetime],
        attr: Optional[str],
        obj: Any,
        **kwargs,
    ):
        if value is None:
            return None
        return DT(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        if value is None:
            return None
        try:
            return DT(value)
        except ValueError:
            return None


class DateDicomField(StringLikeDicomField):
    def _serialize(
        self, value: Optional[datetime.date], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return None
        return DA(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        if value is None:
            return None
        try:
            return DA(value)
        except ValueError:
            return None


class TimeDicomField(StringLikeDicomField):
    def _serialize(
        self, value: Optional[datetime.time], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return None
        return TM(value)

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        if value is None:
            return None
        try:
            return TM(value)
        except ValueError:
            return None


class BooleanDicomField(fields.Boolean):
    def __init__(
        self,
        truthy: Optional[str] = None,
        falsy: Optional[str] = None,
        **kwargs,
    ):
        if truthy is None:
            truthy = "YES"
        if falsy is None:
            falsy = "NO"
        super().__init__(truthy=set([truthy]), falsy=set([falsy]), **kwargs)

    def _serialize(self, value: bool, attr: Optional[str], obj: Any, **kwargs):
        if value:
            string_value = self.truthy
        else:
            string_value = self.falsy
        return list(string_value)[0]

    def _deserialize(self, value, attr, data, **kwargs) -> Any:
        deserialized = super()._deserialize(value, attr, data, **kwargs)
        if deserialized == "":
            return None
        return deserialized


class OffsetInSlideCoordinateSystemField(fields.Field):
    def _serialize(
        self, origin: Optional[PointMm], attr: Optional[str], obj: Any, **kwargs
    ):
        if origin is None:
            if self.dump_default is None:
                return None
            assert isinstance(self.dump_default, PointMm)
            origin = self.dump_default
        origin_element = Dataset()
        origin_element.XOffsetInSlideCoordinateSystem = DSfloat(origin.x, True)
        origin_element.YOffsetInSlideCoordinateSystem = DSfloat(origin.y, True)
        return [origin_element]

    def _deserialize(
        self,
        value: DataElement,
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ) -> PointMm:
        return PointMm(
            x=value[0].XOffsetInSlideCoordinateSystem,
            y=value[0].YOffsetInSlideCoordinateSystem,
        )


class ImageOrientationSlideField(fields.Field):
    def _serialize(
        self, rotation: Optional[float], attr: Optional[str], obj: Any, **kwargs
    ):
        if rotation is None:
            if self.dump_default is None:
                return None
            assert isinstance(self.dump_default, float)
            rotation = self.dump_default
        x = round(math.sin(rotation * math.pi / 180), 8)
        y = round(math.cos(rotation * math.pi / 180), 8)
        return [-x, y, 0, y, x, 0]

    def _deserialize(
        self,
        value: Tuple[float, float, float, float, float, float],
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ) -> float:
        orientation = Orientation(value)
        return orientation.rotation


class ListDicomField(fields.List):
    """Wrapper around normal list that handles single-valued lists from pydicom."""

    def __init__(
        self,
        cls_or_instance: Union[fields.Field, type],
        dump_none_if_empty: bool = False,
        **kwargs,
    ):
        self._dump_none_if_empty = dump_none_if_empty
        super().__init__(cls_or_instance, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Any:
        if self._dump_none_if_empty and value is None or len(value) == 0:
            return None
        return super()._serialize(value, attr, obj, **kwargs)

    def _deserialize(
        self, value: Union[Any, List[Any]], attr, data, **kwargs
    ) -> List[Any]:
        if not isinstance(value, MultiValue):
            value = [value]
        return super()._deserialize(value, attr, data, **kwargs)


class FlattenOnDumpNestedDicomField(fields.Nested):
    """Field that flattens the nested dataset into the parent dataset on dump.

    On load the nested fields are deflatten from the parent dataset to a nested dataset.

    The flatten/deflatten is done by the parent schema.
    """

    def __init__(self, nested: Schema, **kwargs):
        self._nested = nested
        super().__init__(nested=nested, **kwargs)

    @property
    def nested_schema(self) -> Schema:
        return self._nested

    def de_flatten(self, dataset: Dataset) -> Optional[Dataset]:
        """Create new dataset containing the attributes defined in nested schema."""
        nested = Dataset()
        for nested_field in self.nested_schema.fields.values():
            if nested_field.dump_only:
                continue
            if isinstance(nested_field, FlattenOnDumpNestedDicomField):
                de_flatten_nested_field = nested_field.de_flatten(dataset)
                if de_flatten_nested_field is not None:
                    nested.update(de_flatten_nested_field)
            elif nested_field.data_key is not None and nested_field.data_key in dataset:
                nested_value = dataset.get(nested_field.data_key)
                setattr(nested, nested_field.data_key, nested_value)
        if len(nested) == 0:
            return None
        return nested

    def flatten(self, data: Dict[str, Any]):
        """Insert attributes from nested dataset into data."""
        key = self.name
        if self.data_key is not None:
            key = self.data_key
        nested = data.pop(key, None)
        if isinstance(nested, Dataset):
            for nested_key, nested_value in nested.items():
                data[nested_key] = nested_value  # type: ignore

    def _serialize(self, nested_obj, attr: Optional[str], obj: Any, **kwargs):
        if nested_obj is None and self.dump_default != missing:
            nested_obj = self.dump_default
        return super()._serialize(nested_obj, attr, obj, **kwargs)


class FloatDicomField(fields.Float):
    def _serialize(self, value: float, attr: Optional[str], obj: Any, **kwargs):
        return DSfloat(value)


CodeType = TypeVar("CodeType", Code, ConceptCode)


class CodeDicomSchema(Schema):
    """Schema for a DICOM `code sequence`."""

    value = StringDicomField(VR.SH, data_key="CodeValue")
    scheme_designator = StringDicomField(VR.SH, data_key="CodingSchemeDesignator")
    meaning = StringDicomField(VR.LO, data_key="CodeMeaning")
    scheme_version = StringDicomField(
        VR.SH, data_key="CodingSchemeVersion", allow_none=True
    )

    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._load_type = load_type
        super().__init__(**kwargs)

    @pre_dump
    def pre_dump(self, code: Optional[Code] = None, **kwargs):
        if code is None:
            return None
        data = {
            "value": code.value,
            "scheme_designator": code.scheme_designator,
            "meaning": code.meaning,
        }
        if code.scheme_version is not None and code.scheme_version != "":
            data["scheme_version"] = code.scheme_version
        return data

    @post_dump
    def post_dump(self, data: Dict[str, Any], **kwargs):
        dataset = Dataset()
        for data_key in (
            field.data_key
            for field in self.fields.values()
            if field.data_key is not None and field.data_key in data
        ):
            setattr(dataset, data_key, data.get(data_key))
        return dataset

    @pre_load
    def pre_load(self, dataset: Dataset, **kwargs):
        return {
            data_key: dataset.get(data_key)
            for data_key in (
                field.data_key
                for field in self.fields.values()
                if field.data_key is not None
            )
            if dataset.get(data_key) not in (None, "")
        }

    @post_load
    def post_load(self, data: Dict[str, Any], **kwargs):
        return self._load_type(**data)


class CodeDicomField(fields.Nested, Generic[CodeType]):
    """Field for a DICOM `code sequence`."""

    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        super().__init__(nested=CodeDicomSchema(load_type), **kwargs)


class SingleCodeSequenceField(CodeDicomField):
    """Field for a DICOM `code sequence` that can only contain one code."""

    def _serialize(self, value: CodeType, attr: Optional[str], obj: Any, **kwargs):
        return [super()._serialize(value, attr, obj, **kwargs)]

    def _deserialize(
        self,
        value: Sequence[Dataset],
        attr: Optional[str],
        data: Optional[Dict[str, Any]],
        **kwargs,
    ):
        return super()._deserialize(value[0], attr, data, **kwargs)


class FloatOrCodeDicomField(fields.Field, Generic[CodeType]):
    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._float_field = FloatDicomField()
        self._code_field = CodeDicomField(load_type)
        super().__init__(**kwargs)

    def _serialize(
        self, value: Union[float, CodeType], attr: Optional[str], obj: Any, **kwargs
    ):
        assert attr is not None
        if isinstance(value, float):
            return self._float_field.serialize(attr, obj, **kwargs)
        return self._code_field.serialize(attr, obj, **kwargs)

    def _deserialize(self, value: Union[DSfloat, Dataset], attr, data, **kwargs):
        if isinstance(value, DSfloat):
            return self._float_field.deserialize(value, attr, data, **kwargs)
        return self._code_field.deserialize(value, attr, data, **kwargs)


class UidDicomField(StringDicomField):
    def __init__(self, **kwargs):
        super().__init__(value_representation=VR.UI, **kwargs)

    def _deserialize(self, value: Any, attr, data, **kwargs):
        if value is None or value == "":
            return None
        if isinstance(value, UID):
            return value
        return UID(value)


class UidDatasetDicomField(UidDicomField):
    def __init__(self, data_key: str, **kwargs):
        self._data_key = data_key
        super().__init__(data_key=data_key, **kwargs)

    def _serialize(
        self, value: Optional[UID], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        setattr(dataset, self._data_key, value)
        return dataset

    def _deserialize(self, value: Dataset, attr, data, **kwargs) -> Optional[UID]:
        nested_value = getattr(value, self._data_key, None)
        if nested_value is None:
            return None
        return super()._deserialize(nested_value, attr, data, **kwargs)


class PersonNameDicomField(fields.String):
    def _deserialize(self, value: PersonName, attr, data, **kwargs) -> str:
        return str(value)

    def _serialize(self, value: str, attr, obj, **kwargs) -> Optional[PersonName]:
        return PersonName(value)


class IssuerOfIdentifierDicomField(fields.Field):
    def _deserialize(
        self, value: Optional[Sequence[Dataset]], attr, data, **kwargs
    ) -> Optional[IssuerOfIdentifier]:
        if value is None or len(value) == 0:
            return None
        dataset = value[0]
        if "UniversalEntityIDType" in dataset:
            return UniversalIssuerOfIdentifier(
                dataset.UniversalEntityID,
                UniversalIssuerType(dataset.UniversalEntityIDType),
                dataset.get("LocalNamespaceEntityID", None),
            )
        if "LocalNamespaceEntityID" in dataset:
            return LocalIssuerOfIdentifier(dataset.LocalNamespaceEntityID)
        return None

    def _serialize(
        self, value: Optional[IssuerOfIdentifier], attr, obj, **kwargs
    ) -> Optional[Sequence[Dataset]]:
        if value is None:
            return []
        dataset = Dataset()
        if isinstance(value, UniversalIssuerOfIdentifier):
            dataset.UniversalEntityID = value.identifier
            dataset.UniversalEntityIDType = value.issuer_type.name
            if value.local_identifier is not None:
                dataset.LocalNamespaceEntityID = value.local_identifier
        elif isinstance(value, LocalIssuerOfIdentifier):
            dataset.LocalNamespaceEntityID = value.identifier
        else:
            raise NotImplementedError()
        return [dataset]


class PixelSpacingDicomField(fields.Field):
    def _serialize(
        self, value: Optional[SizeMm], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            return None
        return [DSfloat(value.width, True), DSfloat(value.height, True)]

    def _deserialize(
        self, value: Optional[Sequence[DSfloat]], attr, data, **kwargs
    ) -> Optional[SizeMm]:
        if value is None or len(value) == 0:
            return None
        return SizeMm(value[0], value[1])


ValueType = TypeVar("ValueType")


class TypeDicomField(fields.Field, Generic[ValueType]):
    def __init__(self, nested: fields.Field, **kwargs):
        self._nested = nested
        super().__init__(**kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        return self._nested._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value: Any, attr, data, **kwargs):
        de_serialized = self._nested._deserialize(value, attr, data, **kwargs)
        return de_serialized


class DefaultingDicomField(TypeDicomField[ValueType]):
    """Wrapper around a field that should always be present and have a value. Default
    value is constant."""

    def __init__(self, nested: fields.Field, dump_default: ValueType, **kwargs):
        self._dump_default = dump_default
        super().__init__(nested=nested, dump_default=dump_default, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            value = self._dump_default
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingNoneDicomField(DefaultingDicomField[Optional[ValueType]]):
    """Wrapper around a field that should always be present but can be None."""

    def __init__(self, nested: fields.Field, **kwargs):
        super().__init__(nested=nested, dump_default=None, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingTagDicomField(TypeDicomField[ValueType]):
    """Wrapper around a field that should always be present and have a value. Default
    value is taken from object by attribute defined by tag."""

    def __init__(self, nested: fields.Field, tag: str, **kwargs):
        self._tag = tag
        super().__init__(nested=nested, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ):
        if value is None:
            value = getattr(obj, self._tag)
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingListDicomField(fields.List):
    def __init__(self, nested: fields.Field, dump_default: List, **kwargs):
        self._dump_default = dump_default
        super().__init__(cls_or_instance=nested, dump_default=dump_default, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Any:
        if value is None or len(value) == 0:
            value = self.dump_default
        return super()._serialize(value, attr, obj, **kwargs)


class DefaultingListTagDicomField(fields.List):
    def __init__(self, nested: fields.Field, tag: str, **kwargs):
        self._tag = tag
        super().__init__(cls_or_instance=nested, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs) -> Any:
        if value is None or len(value) == 0:
            value = getattr(obj, self._tag)
        return super()._serialize(value, attr, obj, **kwargs)


class NestedDatasetDicomField(fields.Nested, Generic[ValueType]):
    """fields.Field for attribute of a single-item dataset sequence with a nested
    sing-item dataset sequence with the item the nested schema should handle."""

    def __init__(self, nested: Schema, data_key: str, nested_data_key: str, **kwargs):
        self._nested = nested
        self._data_key = data_key
        self._nested_data_key = nested_data_key
        super().__init__(nested=nested, data_key=data_key, **kwargs)

    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[List[Dataset]]:
        nested_value = super()._serialize(value, attr, obj, **kwargs)
        if nested_value is None:
            return None
        dataset = Dataset()
        setattr(dataset, self._nested_data_key, [nested_value])
        return [dataset]

    def _deserialize(
        self, value: Sequence[Dataset], attr, data, **kwargs
    ) -> Optional[ValueType]:
        nested_value = getattr(value[0], self._nested_data_key, None)
        if nested_value is None:
            return None
        return super()._deserialize(
            nested_value[0], attr, data, **kwargs
        )  # type: ignore


class ContentItemDicomField(fields.Field, Generic[ValueType]):
    @abstractmethod
    def _serialize(
        self, value: Optional[ValueType], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        raise NotImplementedError()

    @abstractmethod
    def _deserialize(
        self, value: Dataset, attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[ValueType]:
        raise NotImplementedError()


class CodeItemDicomField(ContentItemDicomField[Code]):
    def __init__(self, load_type: Type[CodeType], **kwargs) -> None:
        self._load_type = load_type
        super().__init__(**kwargs)

    def _serialize(
        self, value: Optional[Code], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        code_dataset = Dataset()
        code_dataset.CodeValue = value.value
        code_dataset.CodingSchemeDesignator = value.scheme_designator
        code_dataset.CodeMeaning = value.meaning
        if value.scheme_version is not None and value.scheme_version != "":
            code_dataset.CodingSchemeVersion = value.scheme_version
        dataset = Dataset()
        dataset.ValueType = "CODE"
        dataset.ConceptCodeSequence = [code_dataset]
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        version = dataset.ConceptCodeSequence[0].get("CodingSchemeVersion", None)
        if version == "":
            version = None
        return self._load_type(
            value=dataset.ConceptCodeSequence[0].CodeValue,
            scheme_designator=dataset.ConceptCodeSequence[0].CodingSchemeDesignator,
            meaning=dataset.ConceptCodeSequence[0].CodeMeaning,
            scheme_version=version,
        )


class StringItemDicomField(ContentItemDicomField[str]):
    def _serialize(
        self, value: Optional[str], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        dataset.ValueType = "TEXT"
        dataset.TextValue = value
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        return dataset.TextValue


class DateTimeItemDicomField(ContentItemDicomField[datetime.datetime]):
    def _serialize(
        self,
        value: Optional[datetime.datetime],
        attr: Optional[str],
        obj: Any,
        **kwargs,
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        dataset.ValueType = "DATETIME"
        dataset.DateTime = DT(value)
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        try:
            return DT(dataset.DateTime)
        except ValueError:
            return None


class MeasurementtemDicomField(ContentItemDicomField[Measurement]):
    def _serialize(
        self, value: Optional[Measurement], attr: Optional[str], obj: Any, **kwargs
    ) -> Optional[Dataset]:
        if value is None:
            return None
        dataset = Dataset()
        dataset.ValueType = "NUMERIC"
        dataset.NumericValue = DSfloat(value.value)
        dataset.FloatingPointValue = value.value
        unit_dataset = Dataset()
        unit_dataset.CodeValue = value.unit.value
        unit_dataset.CodingSchemeDesignator = value.unit.scheme_designator
        unit_dataset.CodeMeaning = value.unit.meaning
        if value.unit.scheme_version is not None and value.unit.scheme_version != "":
            unit_dataset.CodingSchemeVersion = value.unit.scheme_version
        dataset.MeasurementUnitsCodeSequence = [unit_dataset]
        return dataset

    def _deserialize(self, dataset: Dataset, attr: Optional[str], obj: Any, **kwargs):
        if hasattr(dataset, "FloatingPointValue"):
            value = dataset.FloatingPointValue
        else:
            value = DSfloat(dataset.NumericValue)
            assert isinstance(value, float)
        unit_dataset = dataset.MeasurementUnitsCodeSequence[0]
        version = unit_dataset.get("CodingSchemeVersion", None)
        if version == "":
            version = None
        unit = UnitCode(
            value=unit_dataset.CodeValue,
            scheme_designator=unit_dataset.CodingSchemeDesignator,
            meaning=unit_dataset.CodeMeaning,
            scheme_version=version,
        )
        return Measurement(value=value, unit=unit)



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import List, Optional, Sequence, Set, Tuple

from pydicom.dataset import Dataset
from pydicom.sequence import Sequence as DicomSequence

from wsidicom.geometry import Point, Size
from wsidicom.instance.dataset import WsiDataset


class TileIndex(metaclass=ABCMeta):
    """Index for mapping tile position to frame number. Is subclassed into
    FullTileIndex and SparseTileIndex."""

    def __init__(self, datasets: Sequence[WsiDataset]):
        """Create tile index for frames in datasets. Requires equal tile
        size for all tile planes.

        Parameters
        ----------
        datasets: Sequence[Dataset]
            List of datasets containing tiled image data.

        """
        self._datasets = datasets
        self._image_size = datasets[0].image_size
        self._tile_size = datasets[0].tile_size
        self._frame_count = self._read_frame_count_from_datasets(datasets)
        self._optical_paths = self._read_optical_paths_from_datasets(datasets)
        self._tiled_size = self.image_size.ceil_div(self.tile_size)

    def __str__(self) -> str:
        return (
            f"{type(self).__name__} with image size {self.image_size}, "
            f"tile size {self.tile_size}, tiled size {self.tiled_size}, "
            f"optical paths {self.optical_paths}, "
            f"focal planes {self.focal_planes}, "
            f"and frame count {self.frame_count}"
        )

    @property
    @abstractmethod
    def focal_planes(self) -> List[float]:
        """Return list of focal planes in index."""
        raise NotImplementedError()

    @property
    def image_size(self) -> Size:
        """Return image size in pixels."""
        return self._image_size

    @property
    def tile_size(self) -> Size:
        """Return tile size in pixels."""
        return self._tile_size

    @property
    def tiled_size(self) -> Size:
        """Return size of tiling (columns x rows)."""
        return self._tiled_size

    @property
    def frame_count(self) -> int:
        """Return total number of frames in index."""
        return self._frame_count

    @property
    def optical_paths(self) -> List[str]:
        """Return list of optical paths in index."""
        return self._optical_paths

    @abstractmethod
    def pretty_str(self, indent: int = 0, depth: Optional[int] = None) -> str:
        raise NotImplementedError()

    @abstractmethod
    def get_frame_index(self, tile: Point, z: float, path: str) -> int:
        """Abstract method for getting the frame index for a tile"""
        raise NotImplementedError()

    @staticmethod
    def _read_frame_count_from_datasets(datasets: Sequence[Dataset]) -> int:
        """Return total frame count from files.

        Parameters
        ----------
        datasets: Sequence[Dataset]
           List of datasets.

        Returns
        -------
        int
            Total frame count.

        """
        count = 0
        for dataset in datasets:
            count += dataset.frame_count
        return count

    @classmethod
    def _read_optical_paths_from_datasets(
        cls, datasets: Sequence[Dataset]
    ) -> List[str]:
        """Return list of optical path identifiers from files.

        Parameters
        ----------
        datasets: Sequence[Dataset]
           List of datasets.

        Returns
        -------
        List[str]
            Optical identifiers.

        """
        paths: Set[str] = set()
        for dataset in datasets:
            paths.update(cls._get_path_identifers(dataset.optical_path_sequence))
        if len(paths) == 0:
            return ["0"]
        return list(paths)

    @staticmethod
    def _get_path_identifers(
        optical_path_sequence: Optional[DicomSequence],
    ) -> List[str]:
        """Parse optical path sequence and return list of optical path
        identifiers

        Parameters
        ----------
        optical_path_sequence: DicomSequence
            Optical path sequence.

        Returns
        -------
        List[str]
            List of optical path identifiers.
        """
        if optical_path_sequence is None:
            return ["0"]
        return list(
            {
                str(optical_ds.OpticalPathIdentifier)
                for optical_ds in optical_path_sequence
            }
        )

    def _read_frame_coordinates(self, frame: Dataset) -> Tuple[Point, float]:
        """Return frame coordinate (Point(x, y) and float z) of the frame.
        In the Plane Position Slide Sequence x and y are defined in mm and z in
        um.

        Parameters
        ----------
        frame: Dataset
            Pydicom frame sequence.

        Returns
        -------
        Point, float
            The frame xy coordinate and z coordinate
        """
        DECIMALS = 3
        position = frame.PlanePositionSlideSequence[0]
        y = int(position.RowPositionInTotalImagePixelMatrix) - 1
        x = int(position.ColumnPositionInTotalImagePixelMatrix) - 1
        z_offset = getattr(position, "ZOffsetInSlideCoordinateSystem", 0.0)
        z = round(float(z_offset), DECIMALS)
        tile = Point(x=x, y=y) // self.tile_size
        return tile, z



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Json fields for serializing values."""


import dataclasses
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Type, Union

import numpy as np
from marshmallow import ValidationError, fields
from pydicom.sr.coding import Code
from pydicom.uid import UID

from wsidicom.conceptcode import CidConceptCode, CidConceptCodeType, UnitCode
from wsidicom.geometry import PointMm, SizeMm
from wsidicom.metadata.optical_path import LutDataType
from wsidicom.metadata.sample import (
    IssuerOfIdentifier,
    LocalIssuerOfIdentifier,
    Measurement,
    SpecimenIdentifier,
    UniversalIssuerOfIdentifier,
    UniversalIssuerType,
)


class IssuerOfIdentifierJsonField(fields.Field):
    def _serialize(
        self, value: Optional[IssuerOfIdentifier], attr, obj, **kwargs
    ) -> Optional[Union[str, Dict]]:
        if value is None:
            return None
        if isinstance(value, LocalIssuerOfIdentifier):
            return {
                "identifier": value.identifier,
            }
        elif isinstance(value, UniversalIssuerOfIdentifier):
            serialized = {
                "identifier": value.identifier,
                "issuer_type": value.issuer_type.name,
            }
            if value.local_identifier is not None:
                serialized["local_identifier"] = value.local_identifier
            return serialized
        raise NotImplementedError(f"Serialization of {type(value)} is not implemented.")

    def _deserialize(
        self,
        value: Dict[str, Any],
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> IssuerOfIdentifier:
        try:
            identifier = value["identifier"]
            if "issuer_type" in value:
                issuer_type = UniversalIssuerType(value["issuer_type"])
                local_identifier = value.get("local_identifier", None)
                return UniversalIssuerOfIdentifier(
                    identifier, issuer_type, local_identifier
                )
            return LocalIssuerOfIdentifier(identifier)
        except ValueError as error:
            raise ValidationError(
                "Could not deserialize issuer of identifier."
            ) from error


class SpecimenIdentifierJsonField(fields.Field):
    _issuer_of_identifier_field = IssuerOfIdentifierJsonField()

    def _serialize(
        self,
        value: Optional[Union[str, SpecimenIdentifier]],
        attr,
        obj,
        **kwargs,
    ) -> Optional[Union[str, Dict]]:
        if value is None:
            return None
        if isinstance(value, str):
            return value
        if value.issuer is None:
            return {
                "value": value.value,
            }

        return {
            "value": value.value,
            "issuer": self._issuer_of_identifier_field._serialize(
                value.issuer, attr, obj, **kwargs
            ),
        }

    def _deserialize(
        self,
        value: Union[str, Dict],
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> Union[str, SpecimenIdentifier]:
        try:
            if isinstance(value, str):
                return value
            if "issuer" not in value:
                return SpecimenIdentifier(value["value"])
            issuer = self._issuer_of_identifier_field._deserialize(
                value["issuer"], attr, data, **kwargs
            )
            return SpecimenIdentifier(value["value"], issuer)

        except ValueError as error:
            raise ValidationError(
                "Could not deserialize specimen identifier."
            ) from error


class PointMmJsonField(fields.Field):
    def _serialize(
        self, value: Optional[PointMm], attr, obj, **kwargs
    ) -> Optional[Dict]:
        if value is None:
            return None
        return {
            field.name: getattr(value, field.name)
            for field in dataclasses.fields(value)
        }

    def _deserialize(
        self,
        value: Dict,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> PointMm:
        try:
            return PointMm(**value)
        except ValueError as error:
            raise ValidationError("Could not deserialize PointMm.") from error


class SizeMmJsonField(fields.Field):
    def _serialize(
        self, value: Optional[SizeMm], attr, obj, **kwargs
    ) -> Optional[Dict]:
        if value is None:
            return None
        return {
            field.name: getattr(value, field.name)
            for field in dataclasses.fields(value)
        }

    def _deserialize(
        self,
        value: Dict,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> SizeMm:
        try:
            return SizeMm(**value)
        except ValueError as error:
            raise ValidationError("Could not deserialize PointMm.") from error


class UidJsonField(fields.Field):
    def _serialize(self, value: Optional[UID], attr, obj, **kwargs) -> Optional[str]:
        if value is None:
            return None
        return str(value)

    def _deserialize(self, value: str, attr, data, **kwargs) -> UID:
        try:
            return UID(value)
        except ValueError as error:
            raise ValidationError("Could not deserialize UID.") from error


class CodeJsonField(fields.Field):
    def _serialize(self, value: Optional[Code], attr, obj, **kwargs) -> Optional[Dict]:
        if value is None:
            return None
        return JsonFieldFactory._serialize_code(value)

    def _deserialize(self, value: Dict, attr, data, **kwargs) -> Code:
        try:
            return Code(**value)
        except ValueError as error:
            raise ValidationError("Could not deserialize Code.") from error


class StringOrCodeJsonField(fields.Field):
    def _serialize(
        self, value: Optional[Union[str, Code]], attr, obj, **kwargs
    ) -> Optional[Union[str, Dict]]:
        if value is None:
            return None
        if isinstance(value, str):
            return value
        code = {
            "value": value.value,
            "scheme_designator": value.scheme_designator,
            "meaning": value.meaning,
        }
        if value.scheme_version is not None:
            code["scheme_version"] = value.scheme_version
        return code

    def _deserialize(
        self, value: Union[str, Dict], attr, data, **kwargs
    ) -> Union[str, Code]:
        if isinstance(value, str):
            return value
        try:
            return Code(**value)
        except ValueError as error:
            raise ValidationError("Could not deserialize Code.") from error


class JsonFieldFactory:
    @classmethod
    def float_or_concept_code(
        cls, concept_code_type: Type[CidConceptCodeType], many=False, **metadata
    ) -> Type[fields.Field]:
        def serialize(
            self, value: Optional[Union[float, CidConceptCodeType]], attr, obj, **kwargs
        ) -> Optional[Union[float, Dict]]:
            if isinstance(value, float):
                return value
            if isinstance(value, (Code, CidConceptCode)):
                return cls._serialize_code(value)

        def deserialize(
            self,
            value: Union[float, Dict[str, Any]],
            attr: Optional[str],
            data: Optional[Mapping[str, Any]],
            **kwargs,
        ) -> Union[float, CidConceptCodeType]:
            if isinstance(value, (int, float)):
                return value
            try:
                return concept_code_type(**value)
            except (ValueError, KeyError) as error:
                raise ValidationError("Could not deserialize Code.") from error

        return type(
            f"FloatOr{CidConceptCodeType}Field",
            (fields.Field,),
            {"_serialize": serialize, "_deserialize": deserialize},
        )

    @classmethod
    def str_or_concept_code(
        cls, concept_code_type: Type[CidConceptCodeType], many=False, **metadata
    ) -> Type[fields.Field]:
        def serialize(
            self, value: Optional[Union[str, CidConceptCodeType]], attr, obj, **kwargs
        ) -> Optional[Union[str, Dict]]:
            if value is None:
                return None
            if isinstance(value, str):
                return value

            return cls._serialize_code(value)

        def deserialize(
            self,
            value: Union[str, Dict],
            attr: Optional[str],
            data: Optional[Mapping[str, Any]],
            **kwargs,
        ) -> Union[str, CidConceptCodeType]:
            if isinstance(value, str):
                return value
            try:
                return concept_code_type(**value)
            except ValueError as error:
                raise ValidationError("Could not deserialize Code.") from error

        return type(
            f"StringOr{CidConceptCodeType}Field",
            (fields.Field,),
            {"_serialize": serialize, "_deserialize": deserialize},
        )

    @classmethod
    def concept_code(
        cls,
        concept_code_type: Type[CidConceptCodeType],
    ) -> Type[fields.Field]:
        def serialize(
            self, value: Optional[CidConceptCodeType], attr, obj, **kwargs
        ) -> Optional[Dict]:
            if value is None:
                return None
            return cls._serialize_code(value)

        deserialize = cls._concept_code_deserializer_factory(concept_code_type)

        return type(
            f"{CidConceptCodeType}Field",
            (fields.Field,),
            {"_serialize": serialize, "_deserialize": deserialize},
        )

    @staticmethod
    def _concept_code_deserializer_factory(
        concept_code_type: Type[CidConceptCodeType],
    ):
        def _deserialize(
            self,
            value: Dict,
            attr: Optional[str],
            data: Optional[Mapping[str, Any]],
            **kwargs,
        ) -> CidConceptCodeType:
            try:
                return concept_code_type(**value)
            except ValueError as error:
                raise ValidationError("Could not deserialize Code.") from error

        return _deserialize

    @staticmethod
    def _serialize_code(code: Union[Code, CidConceptCode]) -> Dict[str, str]:
        try:
            result = {
                "value": code.value,
                "scheme_designator": code.scheme_designator,
                "meaning": code.meaning,
            }
            if code.scheme_version is not None:
                result["scheme_version"] = code.scheme_version
            return result

        except Exception as exception:
            raise ValueError(f"Failed to serialize code {code}") from exception


class NpUIntDTypeField(fields.Field):
    def _serialize(
        self, value: LutDataType, attr: Optional[str], obj: Any, **kwargs
    ) -> int:
        return 8 * value().itemsize

    def _deserialize(
        self,
        value: int,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> LutDataType:
        if value == 8:
            return np.uint8
        if value == 16:
            return np.uint16
        raise NotImplementedError(f"Not-implemented bit count {value}.")


class FileLoadingField(fields.Field):
    def _serialize(self, value: bytes, attr: Optional[str], obj: Any, **kwargs):
        raise NotImplementedError("Dumping bytes to file not implemented.")

    def _deserialize(
        self,
        value: str,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        path = Path(value)
        if not path.exists() or not path.is_file():
            raise ValidationError(f"File {path} does not exist or is not a file.")
        with open(path, "rb") as file:
            return file.read()


class MeasurementJsonField(fields.Field):
    def _serialize(
        self, value: Optional[Measurement], attr, obj, **kwargs
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return None
        return {"value": value.value, "unit": value.unit.value}

    def _deserialize(
        self,
        value: Dict[str, Any],
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> Measurement:
        try:
            return Measurement(value["value"], UnitCode.from_unit(value["unit"]))
        except ValueError as error:
            raise ValidationError("Could not deserialize measurement.") from error



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Slide model."""

from dataclasses import dataclass
from typing import Optional, Sequence, Union

from wsidicom.metadata.sample import SlideSample, SpecimenIdentifier, Staining


@dataclass(frozen=True)
class Slide:
    """
    Metadata for a slide.

    A slide has a an identifier and contains one or more samples. All the samples on the
    slide has been stained with the same stainings.

    Corresponds to attributes in the Specimen Module:
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.7.6.22.html

    Parameters
    ----------
    identifier : Optional[Union[str, SpecimenIdentifier]] = None
        The identifier of the slide (Container Identifier).
    stainings : Optional[Sequence[Staining]] = None
        List of stainings used on the slide.
    samples : Optional[Sequence[SlideSample]] = None
        List of samples on the slide.
    """

    identifier: Optional[Union[str, SpecimenIdentifier]] = None
    stainings: Optional[Sequence[Staining]] = None
    samples: Optional[Sequence[SlideSample]] = None



#    Copyright 2024 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module for opening WsiDicomIO instances from streams or files."""

from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Dict,
    Iterable,
    Iterator,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from fsspec.core import url_to_fs
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem
from pydicom.uid import UID
from upath import UPath

from wsidicom.file.io.wsidicom_io import WsiDicomIO


class WsiDicomStreamOpener:
    def __init__(
        self,
        file_options: Optional[Dict[str, Any]] = None,
    ):
        """Create a WsiDicomStreamOpener.

        Parameters
        ----------
        file_options: Optional[Dict[str, Any]] = None
            Keyword arguments for opening filesystems and files.
        """

        self._file_options = file_options or {}

    def open(
        self,
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]],
        sop_class_uids: Optional[Union[UID, Sequence[UID]]] = None,
    ) -> Iterator[WsiDicomIO]:
        """Open DICOM streams in paths and return WsiDicomIO instances.

        Parameters
        ----------
        files: Union[str, Path, UPath, Iterable[Union[str, Path, UPath]]],
            Folder, single file, or sequence of files to open.
        sop_class_uids: Optional[Union[UID, Sequence[UID]]] = None,
            SOP class uids to filter on.

        Returns
        -------
        Iterator[WsiDicomIO]
            Opened WsiDicomIO instances.
        """
        if isinstance(sop_class_uids, UID):
            sop_class_uids = [sop_class_uids]
        if isinstance(files, (str, Path, UPath)):
            files = [files]
        for file in files:
            for stream in self._open_streams(str(file), "rb"):
                try:
                    if hasattr(stream, "path"):
                        path = UPath(getattr(stream, "path"))
                    else:
                        path = None
                    dicom_io = WsiDicomIO(stream, owned=True, filepath=path)
                    if dicom_io.is_dicom and (
                        sop_class_uids is None
                        or dicom_io.media_storage_sop_class_uid in sop_class_uids
                    ):
                        yield dicom_io
                    else:
                        stream.close()
                except Exception:
                    stream.close()

    def open_for_writing(
        self,
        path: Union[str, Path, UPath],
        mode: Union[Literal["r+b"], Literal["w+b"]],
    ) -> WsiDicomIO:
        """Open a stream for writing.

        Parameters
        ----------
        path: Union[str, Path, UPath]
            Path to open.
        mode: Union[Literal["r+b"], Literal["w+b"]]
            Mode to open in.

        Returns
        -------
        WsiDicomIO
            Opened WsiDicomIO instance.
        """
        fs, path = self._open_filesystem(str(path))
        fs.makedirs(UPath(path).parent, exist_ok=True)
        stream = self._open_stream(fs, path, mode)
        return WsiDicomIO(stream, owned=True, filepath=UPath(path))

    def _open_streams(
        self,
        path: str,
        mode: Union[Literal["rb"], Literal["r+b"], Literal["w+b"]],
    ) -> Iterator[Union[BinaryIO, AbstractBufferedFile]]:
        """Open streams from path. If path is a directory, open all files in directory.

        Parameters
        ----------
        path: str
            Path to open.
        mode: Union[Literal["rb"], Literal["r+b"], Literal["w+b"]]
            Mode to open in.

        Returns
        -------
        Iterator[Union[BinaryIO, AbstractBufferedFile]]
            Opened streams.
        """
        fs, path = self._open_filesystem(path)
        if fs.isdir(str(path)):
            files = (file for file in fs.ls(str(path), detail=False) if fs.isfile(file))
        elif fs.isfile(str(path)):
            files = [str(path)]
        else:
            files = (
                file
                for file in fs.glob(str(path))
                if isinstance(file, str) and fs.isfile(file)
            )
        for file in files:
            yield self._open_stream(fs, file, mode)

    def _open_filesystem(self, path: str) -> Tuple[AbstractFileSystem, str]:
        """Open fsspec filesystem from path.

        Parameters
        ----------
        path: str
            Path to open.

        Returns
        -------
        Tuple[AbstractFileSystem, str]
            Opened filesystem and path.
        """
        fs, path = url_to_fs(path, **self._file_options or {})
        return fs, path  # type: ignore

    def _open_stream(
        self,
        fs: AbstractFileSystem,
        path: str,
        mode: Union[Literal["rb"], Literal["r+b"], Literal["w+b"]],
    ) -> AbstractBufferedFile:
        """Open stream from path.

        Parameters
        ----------
        fs: AbstractFileSystem
            Filesystem to open from.
        path: str
            Path to open.
        mode: Union[Literal["rb"], Literal["r+b"], Literal["w+b"]]
            Mode to open in.

        Returns
        -------
        AbstractBufferedFile
            Opened stream.
        """
        return fs.open(path, mode, **self._file_options or {})  # type: ignore



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with codec supporting different formats."""

from typing import Iterable, Optional, Union

import numpy as np
from PIL.Image import Image
from pydicom.uid import (
    UID,
    AllTransferSyntaxes,
)

from wsidicom.codec.decoder import Decoder
from wsidicom.codec.encoder import Encoder
from wsidicom.geometry import Size


class Codec:
    def __init__(self, decoder: Decoder, encoder: Encoder):
        """Create codec with the given decoder and encoder."""
        self._decoder = decoder
        self._encoder = encoder

    @property
    def encoder(self) -> Encoder:
        return self._encoder

    @property
    def decoder(self) -> Decoder:
        return self._decoder

    @classmethod
    def supported_transfer_syntaxes(
        cls,
        samples_per_pixel: int,
        bits: int,
        photometric_interpretation: str,
        possible_transfer_syntaxes: Optional[Iterable[UID]] = None,
    ) -> Iterable[UID]:
        """Return supported transfer syntaxes for the given parameters.

        Parameters
        ----------
        samples_per_pixel: int
            Samples per pixel of the image.
        bits: int
            Bits per sample of the image.
        photometric_interpretation: str
            Photometric interpretation of the image.
        possible_transfer_syntaxes: Optional[Iterable[UID]]
            Possible transfer syntaxes to check for support.

        Returns
        -------
        Iterable[UID]
            Supported transfer syntaxes for the given parameters.
        """

        if possible_transfer_syntaxes is None:
            possible_transfer_syntaxes = AllTransferSyntaxes
        return (
            transfer_syntax
            for transfer_syntax in AllTransferSyntaxes
            if cls.is_supported(
                transfer_syntax,
                samples_per_pixel,
                bits,
                photometric_interpretation,
            )
        )

    def decode(self, data: bytes) -> Image:
        return self.decoder.decode(data)

    def encode(self, image: Union[Image, np.ndarray]) -> bytes:
        return self.encoder.encode(image)

    @classmethod
    def create(
        cls,
        transfer_syntax: UID,
        samples_per_pixel: int,
        bits: int,
        size: Size,
        photometric_interpretation: str,
    ) -> "Codec":
        """Create codec for the given parameters.

        Parameters
        ----------
        transfer_syntax: UID
            Transfer syntax of the image.
        samples_per_pixel: int
            Samples per pixel of the image.
        bits: int
            Bits per sample of the image.
        size: Size
            Size of the image.
        photometric_interpretation: str
            Photometric interpretation of the image.

        Returns
        -------
        Codec
            Codec for the given parameters.
        """
        decoder = Decoder.create(
            transfer_syntax,
            samples_per_pixel,
            bits,
            size,
            photometric_interpretation,
        )
        encoder = Encoder.create(
            transfer_syntax,
            bits,
            photometric_interpretation,
        )
        return cls(decoder, encoder)

    @classmethod
    def is_supported(
        cls,
        transfer_syntax: UID,
        samples_per_pixel: int,
        bits: int,
        photometric_interpretation: str,
    ) -> bool:
        """Return True if codec supports the given parameters.

        Parameters
        ----------
        transfer_syntax: UID
            Transfer syntax of the image.
        samples_per_pixel: int
            Samples per pixel of the image.
        bits: int
            Bits per sample of the image.
        photometric_interpretation: str
            Photometric interpretation of the image.

        Returns
        -------
        bool
            True if codec supports the given parameters.
        """
        decoder = Decoder.select_decoder(transfer_syntax, samples_per_pixel, bits)
        if decoder is None:
            return False
        try:
            Encoder.create(
                transfer_syntax,
                bits,
                photometric_interpretation,
            )
        except Exception:
            return False
        return True



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with metadata models and schemas for serialization to and from DICOM or
json."""

from wsidicom.metadata.equipment import Equipment
from wsidicom.metadata.image import (
    ExtendedDepthOfField,
    FocusMethod,
    Image,
    ImageCoordinateSystem,
)
from wsidicom.metadata.label import Label
from wsidicom.metadata.optical_path import (
    ConstantLutSegment,
    DiscreteLutSegment,
    ImagePathFilter,
    LightPathFilter,
    LinearLutSegment,
    Lut,
    Objectives,
    OpticalPath,
)
from wsidicom.metadata.patient import Patient, PatientDeIdentification, PatientSex
from wsidicom.metadata.sample import (
    Collection,
    Embedding,
    Fixation,
    LocalIssuerOfIdentifier,
    Measurement,
    Processing,
    Receiving,
    Sample,
    SampleLocalization,
    Sampling,
    SamplingLocation,
    SlideSample,
    Specimen,
    SpecimenIdentifier,
    Staining,
    Storage,
    UniversalIssuerOfIdentifier,
    UniversalIssuerType,
    UnknownSampling,
)
from wsidicom.metadata.series import Series
from wsidicom.metadata.slide import Slide
from wsidicom.metadata.study import Study
from wsidicom.metadata.wsi import WsiMetadata

__all__ = [
    "Equipment",
    "ExtendedDepthOfField",
    "FocusMethod",
    "Image",
    "ImageCoordinateSystem",
    "Label",
    "ConstantLutSegment",
    "DiscreteLutSegment",
    "ImagePathFilter",
    "LightPathFilter",
    "LinearLutSegment",
    "Lut",
    "Objectives",
    "OpticalPath",
    "Patient",
    "PatientDeIdentification",
    "PatientSex",
    "Collection",
    "Embedding",
    "Fixation",
    "LocalIssuerOfIdentifier",
    "Measurement",
    "Processing",
    "Receiving",
    "Sample",
    "SampleLocalization",
    "Sampling",
    "SamplingLocation",
    "SlideSample",
    "Specimen",
    "SpecimenIdentifier",
    "Staining",
    "Storage",
    "UniversalIssuerOfIdentifier",
    "UniversalIssuerType",
    "UnknownSampling",
    "Series",
    "Slide",
    "Study",
    "WsiMetadata",
]



#    Copyright 2021, 2022, 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Class for reading frame positions from non-encapsulated data."""

from typing import List, Tuple

from wsidicom.file.io.frame_index.frame_index import FrameIndex
from wsidicom.file.io.frame_index.offset_table_type import OffsetTableType
from wsidicom.file.io.wsidicom_io import WsiDicomIO
from wsidicom.geometry import Size


class NativePixelData(FrameIndex):
    def __init__(
        self,
        file: WsiDicomIO,
        pixel_data_start: int,
        frame_count: int,
        tile_size: Size,
        samples_per_pixel: int,
        bits: int,
    ):
        self._tile_size = tile_size
        self._samples_per_pixel = samples_per_pixel
        self._bits = bits
        super().__init__(file, pixel_data_start, frame_count)

    @property
    def offset_table_type(self) -> OffsetTableType:
        return OffsetTableType.NONE

    def _get_pixels_start(self) -> int:
        self._validate_pixel_data_start()
        return self._file.tell()

    def _get_index(self) -> List[Tuple[int, int]]:
        """Create frame positions for uncapsulated data.

        Parameters
        ----------
        pixel_data_start: int
            Offset to first frame in pixel data.

        Returns
        -------
        List[Tuple[int, int]]
            A list with frame positions and frame lengths.
        """
        frame_size = self._tile_size.area * self._samples_per_pixel * (self._bits // 8)
        return [
            (self._pixels_start + index * frame_size, frame_size)
            for index in range(self._frame_count)
        ]

    def _validate_pixel_data_start(self):
        """Check that pixel data tag is present and that the tag length is equal to
        expected count. Raises WsiDicomFileError otherwise.

        """
        expected_length = (
            self._tile_size.area
            * self._samples_per_pixel
            * (self._bits // 8)
            * self._frame_count
        )
        super()._validate_pixel_data_start(expected_length)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Module with decoders for image data."""

import io
from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, Optional, Type

import numpy as np
from PIL import Image as Pillow
from PIL.Image import Image
from pydicom import config as pydicom_config
from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.encaps import encapsulate
from pydicom.uid import (
    JPEG2000,
    UID,
    JPEG2000Lossless,
    JPEGBaseline8Bit,
    JPEGExtended12Bit,
    JPEGLosslessP14,
    JPEGLosslessSV1,
    JPEGLSLossless,
    JPEGLSNearLossless,
    RLELossless,
)

from wsidicom import config
from wsidicom.codec.optionals import (
    IMAGE_CODECS_AVAILABLE,
    JPEG2K,
    JPEG8,
    JPEGLS,
    PYLIBJPEGLS_AVAILABLE,
    PYLIBJPEGOPENJPEG_AVAILABLE,
    PYLIBJPEGRLE_AVAILABLE,
    jpeg2k_decode,
    jpeg8_decode,
    jpegls_decode,
    pylibjpeg_ls_decode,
    pylibjpeg_openjpeg_decode,
    rle_decode_frame,
)
from wsidicom.codec.rle import RleCodec
from wsidicom.geometry import Size
from wsidicom.uid import HTJPEG2000, HTJPEG2000Lossless, HTJPEG2000RPCLLossless


class Decoder(metaclass=ABCMeta):
    @abstractmethod
    def decode(self, frame: bytes) -> Image:
        """Decode frame into Pillow Image.

        Parameters
        ----------
        frame : bytes
            Encoded frame.

        Returns
        -------
        PIL.Image
            Pillow Image.
        """
        raise NotImplementedError()

    @classmethod
    def is_available(cls) -> bool:
        """Return true if decoder is available.

        Returns
        -------
        bool
            True if decoder is available.
        """
        raise NotImplementedError()

    @classmethod
    @abstractmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        """Return true if decoder supports transfer syntax.

        Parameters
        ----------
        transfer_syntax : pydicom.uid.UID
            Transfer syntax.
        samples_per_pixel : int
            Number of samples per pixel.
        bits : int
            Number of bits per sample.

        Returns
        -------
        bool
            True if decoder supports transfer syntax.

        """
        raise NotImplementedError()

    @classmethod
    def create(
        cls,
        transfer_syntax: UID,
        samples_per_pixel: int,
        bits: int,
        size: Size,
        photometric_interpretation: str,
    ) -> "Decoder":
        """Create a decoder that supports the transfer syntax.

        Parameters
        ----------
        transfer_syntax : UID
            Transfer syntax.
        samples_per_pixel : int
            Number of samples per pixel.
        bits : int
            Number of bits per sample.
        size : Size
            Size of image.
        photometric_interpretation : str
            Photometric interpretation.

        Returns
        -------
        Decoder
            Decoder for transfer syntax.

        """
        if bits != 8 and samples_per_pixel != 1:
            # Pillow only supports 8 bit color images
            raise ValueError(
                f"Non-supported combination of bits {bits} and  "
                f"samples per pixel {samples_per_pixel}. "
                "Non-8 bit images are only supported for grayscale images."
            )
        decoder = cls.select_decoder(transfer_syntax, samples_per_pixel, bits)
        if decoder is None:
            raise ValueError(f"Unsupported transfer syntax: {transfer_syntax}.")
        if decoder == PillowDecoder:
            return PillowDecoder()
        elif decoder == ImageCodecsDecoder:
            return ImageCodecsDecoder(transfer_syntax)
        elif decoder == PylibjpegRleDecoder:
            return PylibjpegRleDecoder(size, samples_per_pixel, bits)
        elif decoder == ImageCodecsRleDecoder:
            return ImageCodecsRleDecoder(size, samples_per_pixel, bits)
        elif decoder == PydicomDecoder:
            return PydicomDecoder(
                transfer_syntax=transfer_syntax,
                size=size,
                samples_per_pixel=samples_per_pixel,
                bits_allocated=(bits // 8) * 8,
                bits_stored=bits,
                photometric_interpretation=photometric_interpretation,
            )
        elif decoder == PyJpegLsDecoder:
            return PyJpegLsDecoder()
        elif decoder == PyLibJpegOpenJpegDecoder:
            return PyLibJpegOpenJpegDecoder()
        raise ValueError(f"Unsupported transfer syntax: {transfer_syntax}")

    @classmethod
    def select_decoder(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> Optional[Type["Decoder"]]:
        """Select decoder based on transfer syntax.

        Parameters
        ----------
        transfer_syntax : UID
            Transfer syntax.
        samples_per_pixel : int
            Number of samples per pixel.
        bits : int
            Number of bits per sample.

        Returns
        -------
        Optional[Type[Decoder]]
            Decoder class. None if no decoder supports transfer syntax.

        """
        if bits != 8 and samples_per_pixel != 1:
            # Pillow only supports 8 bit color images
            return None
        decoders: Dict[str, Type[Decoder]] = {
            "pillow": PillowDecoder,
            "imagecodecs": ImageCodecsDecoder,
            "pylibjpeg_rle": PylibjpegRleDecoder,
            "imagecodecs_rle": ImageCodecsRleDecoder,
            "pylibjpeg_ls": PyJpegLsDecoder,
            "pylibjpeg_openjpeg": PyLibJpegOpenJpegDecoder,
            "pydicom": PydicomDecoder,
        }
        if config.settings.prefered_decoder is not None:
            if config.settings.prefered_decoder not in decoders:
                raise ValueError(
                    f"Unknown preferred decoder: {config.settings.prefered_decoder}."
                )
            decoder = decoders[config.settings.prefered_decoder]
            if not decoder.is_available() or not decoder.is_supported(
                transfer_syntax, samples_per_pixel, bits
            ):
                return None
            return decoder
        return next(
            (
                decoder
                for decoder in decoders.values()
                if decoder.is_available()
                and decoder.is_supported(transfer_syntax, samples_per_pixel, bits)
            ),
            None,
        )

    @staticmethod
    def _set_mode(image: Image) -> Image:
        """Convert image to mode that is fully supported by Pillow."""
        if image.mode.startswith("I;16"):
            # Pillow does not support scaling of 16 bit images
            return image.convert("I")
        return image


class PillowDecoder(Decoder):
    """Decoder that uses Pillow to decode images."""

    _supported_transfer_syntaxes = [
        JPEGBaseline8Bit,
        JPEG2000,
        JPEG2000Lossless,
        HTJPEG2000,
        HTJPEG2000Lossless,
        HTJPEG2000RPCLLossless,
    ]

    def decode(self, frame: bytes) -> Image:
        image = Pillow.open(io.BytesIO(frame))
        return self._set_mode(image)

    @classmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        if bits not in [8, 16]:
            # Pillow only supports 8 and 16 bit images
            return False
        if bits != 8 and samples_per_pixel != 1:
            # Pillow only supports 8 bit color images
            return False
        return transfer_syntax in cls._supported_transfer_syntaxes

    @classmethod
    def is_available(cls) -> bool:
        return True


class NumpyBasedDecoder(Decoder):
    """Base for decoder that returns numpy array."""

    def decode(self, frame: bytes) -> Image:
        decoded = self._decode(frame)
        image = Pillow.fromarray(decoded)
        return self._set_mode(image)

    @abstractmethod
    def _decode(self, frame: bytes) -> np.ndarray:
        """Decode frame into numpy array.

        Parameters
        ----------
        frame : bytes
            Encoded frame.

        Returns
        -------
        np.ndarray
            Numpy array of decoded image."""
        raise NotImplementedError()


class PydicomDecoder(NumpyBasedDecoder):
    """Decoder that uses pydicom to decode images."""

    def __init__(
        self,
        transfer_syntax: UID,
        size: Size,
        samples_per_pixel: int,
        bits_allocated: int,
        bits_stored: int,
        photometric_interpretation: str,
    ):
        """Initialize decoder.

        Parameters
        ----------
        transfer_syntax : UID
            Transfer syntax.
        size : Size
            Size of image.
        samples_per_pixel : int
            Number of samples per pixel.
        bits_allocated : int
            Number of bits allocated.
        bits_stored : int
            Number of bits stored.
        photometric_interpretation : str
            Photometric interpretation.

        """
        handler = self._get_handler(transfer_syntax)
        if handler is None:
            raise ValueError(f"Unsupported transfer syntax: {transfer_syntax}.")
        self._handler = handler
        self._transfer_syntax = transfer_syntax
        dataset = Dataset()
        dataset.file_meta = FileMetaDataset()
        dataset.file_meta.TransferSyntaxUID = transfer_syntax
        dataset.Rows = size.height
        dataset.Columns = size.width
        dataset.SamplesPerPixel = samples_per_pixel
        dataset.BitsAllocated = bits_allocated
        dataset.BitsStored = bits_stored
        dataset.HighBit = bits_stored - 1
        dataset.PhotometricInterpretation = photometric_interpretation
        dataset.PixelRepresentation = 0
        if samples_per_pixel == 3:
            self._reshape_size = (size.height, size.width, samples_per_pixel)
            dataset.PlanarConfiguration = 0
        else:
            self._reshape_size = (size.height, size.width)
        self._dataset = dataset

    def _decode(self, frame: bytes) -> np.ndarray:
        if self._transfer_syntax.is_encapsulated:
            self._dataset.PixelData = encapsulate(frames=[frame])
        else:
            self._dataset.PixelData = frame
        return self._handler(self._dataset).reshape(self._reshape_size)

    @classmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        handler = cls._get_handler(transfer_syntax)
        return handler is not None

    @classmethod
    def _get_handler(
        cls, transfer_syntax: UID
    ) -> Optional[Callable[[Dataset], np.ndarray]]:
        """Return pydicom pixel data handler that supports transfer syntax.

        A pydicom pixel data handler should implement methods `Is_available()`,
        `supports_transfer_syntax()` that takes a transfer syntax and `get_pixeldata()`
        that takes a dataset.
        """
        available_handlers = (
            handler
            for handler in pydicom_config.pixel_data_handlers
            if handler.is_available()
        )
        return next(
            (
                handler.get_pixeldata
                for handler in available_handlers
                if handler.supports_transfer_syntax(transfer_syntax)
            ),
            None,
        )

    @classmethod
    def is_available(cls) -> bool:
        return True


class ImageCodecsDecoder(NumpyBasedDecoder):
    """Decoder that uses imagecodecs to decode images."""

    _supported_transfer_syntaxes = {
        JPEGBaseline8Bit: (jpeg8_decode, JPEG8),
        JPEGExtended12Bit: (jpeg8_decode, JPEG8),
        JPEGLosslessP14: (jpeg8_decode, JPEG8),
        JPEGLosslessSV1: (jpeg8_decode, JPEG8),
        JPEGLSLossless: (jpegls_decode, JPEGLS),
        JPEGLSNearLossless: (jpegls_decode, JPEGLS),
        JPEG2000Lossless: (jpeg2k_decode, JPEG2K),
        JPEG2000: (jpeg2k_decode, JPEG2K),
        HTJPEG2000: (jpeg2k_decode, JPEG2K),
        HTJPEG2000Lossless: (jpeg2k_decode, JPEG2K),
        HTJPEG2000RPCLLossless: (jpeg2k_decode, JPEG2K),
    }

    def __init__(self, transfer_syntax: UID) -> None:
        """Initialize decoder.

        Parameters
        ----------
        transfer_syntax : UID
            Transfer syntax.
        """
        decoder = self._get_decoder(transfer_syntax)
        if decoder is None:
            raise ValueError(f"Unsupported transfer syntax: {transfer_syntax}.")
        self._decoder = decoder

    def _decode(self, frame: bytes) -> np.ndarray:
        if not self.is_available():
            raise RuntimeError("Image codecs not available.")
        return self._decoder(frame)

    @classmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        if not cls.is_available():
            return False
        decoder = cls._get_decoder(transfer_syntax)
        return decoder is not None

    @classmethod
    def is_available(cls) -> bool:
        return IMAGE_CODECS_AVAILABLE

    @classmethod
    def _get_decoder(
        cls, transfer_syntax: UID
    ) -> Optional[Callable[[bytes], np.ndarray]]:
        """Get imagecodes decoder for transfer syntax.

        Parameters
        ----------
        transfer_syntax : UID
            Transfer syntax.

        Returns
        -------
        Optional[Callable[[bytes], np.ndarray]]
            Decoder. None if no decoder supports transfer syntax.
        """
        if transfer_syntax not in cls._supported_transfer_syntaxes:
            return None
        decoder, codec = cls._supported_transfer_syntaxes[transfer_syntax]
        assert decoder is not None and codec is not None
        if not codec.available:
            return None
        return decoder


class RleDecoder(NumpyBasedDecoder):
    def __init__(self, size: Size, samples_per_pixel: int, bits: int):
        if not self.is_supported(RLELossless, samples_per_pixel, bits):
            raise ValueError(
                f"Unsupported combination of samples per pixel {samples_per_pixel} "
                f"and bits {bits}."
            )
        if bits == 8:
            self._dtype = np.uint8
        else:
            self._dtype = np.uint16
        self._size = size
        self._bits = bits
        self._samples_per_pixel = samples_per_pixel
        if samples_per_pixel == 3:
            self._reshape_size = (samples_per_pixel, size.height, size.width)
        else:
            self._reshape_size = (size.height, size.width)

    @classmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        if transfer_syntax != RLELossless:
            return False
        if samples_per_pixel == 3:
            return bits == 8
        if samples_per_pixel == 1:
            return bits in [8, 16]
        return False


class ImageCodecsRleDecoder(RleDecoder):
    @classmethod
    def is_available(cls):
        return IMAGE_CODECS_AVAILABLE

    def _decode(self, frame: bytes) -> np.ndarray:
        if not self.is_available():
            raise RuntimeError("Image codecs not available.")
        return RleCodec.decode(frame, self._size.height, self._size.width, self._bits)


class PylibjpegRleDecoder(RleDecoder):
    """Decoder that uses pylibjpeg-rle to decode images."""

    @classmethod
    def is_available(cls) -> bool:
        return PYLIBJPEGRLE_AVAILABLE

    def _decode(self, frame: bytes) -> np.ndarray:
        if not self.is_available():
            raise RuntimeError("Pylibjpeg-rle not available.")
        decoded_data: bytes = rle_decode_frame(frame, self._size.area, self._bits, "<")
        decoded = np.frombuffer(decoded_data, dtype=self._dtype).reshape(
            self._reshape_size
        )
        if self._samples_per_pixel == 3:
            decoded = decoded.transpose((1, 2, 0))
        return decoded


class PyJpegLsDecoder(NumpyBasedDecoder):
    """Decoder that uses pyjpegls to decode images."""

    @classmethod
    def is_available(cls) -> bool:
        return PYLIBJPEGLS_AVAILABLE

    def _decode(self, frame: bytes) -> np.ndarray:
        if not self.is_available():
            raise RuntimeError("Pylibjpeg not available.")
        return pylibjpeg_ls_decode(np.frombuffer(frame, dtype=np.uint8))

    @classmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        if not cls.is_available():
            return False
        return transfer_syntax in [JPEGLSNearLossless, JPEGLSLossless]


class PyLibJpegOpenJpegDecoder(NumpyBasedDecoder):
    """Decoder that uses pylibjpeg-openjpeg to decode images."""

    @classmethod
    def is_available(cls) -> bool:
        return PYLIBJPEGOPENJPEG_AVAILABLE

    def _decode(self, frame: bytes) -> np.ndarray:
        if not self.is_available():
            raise RuntimeError("Pylibjpeg-openjpeg not available.")
        return pylibjpeg_openjpeg_decode(frame)

    @classmethod
    def is_supported(
        cls, transfer_syntax: UID, samples_per_pixel: int, bits: int
    ) -> bool:
        if not cls.is_available():
            return False
        return transfer_syntax in [
            JPEG2000,
            JPEG2000Lossless,
            HTJPEG2000,
            HTJPEG2000Lossless,
            HTJPEG2000RPCLLossless,
        ]



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Complete WSI model."""

from dataclasses import dataclass, replace
from functools import cached_property
from typing import Optional, Sequence

from pydicom.uid import UID, generate_uid

from wsidicom.metadata.equipment import Equipment
from wsidicom.metadata.image import Image
from wsidicom.metadata.label import Label
from wsidicom.metadata.optical_path import OpticalPath
from wsidicom.metadata.patient import Patient
from wsidicom.metadata.series import Series
from wsidicom.metadata.slide import Slide
from wsidicom.metadata.study import Study


@dataclass(frozen=True)
class WsiMetadata:
    """Metadata for a whole slide image.

    Parameters
    ----------
    study: Study
        Metadata of the study the image belongs to.
    series: Series
        Metadata of the series the image belongs to.
    patient: Patient
        Metadata of the patient the related to the imaged slide.
    equipment: Equipment
        Metadata of the scanner equipment used to acquire the image.
    optical_paths: Sequence[OpticalPath]
        Metadata of the optical paths used to acquire the image.
    slide: Slide
        Metadata of the imaged slide.
    label: Label
        Metadata of the label of the slide.
    image: Image
        Technical metadata of the image.
    frame_of_reference_uid: Optional[UID] = None
        The frame of reference uid of the image.
    dimension_organization_uids: Optional[Sequence[UID]] = None
        The dimension organization uids of the image.
    """

    study: Study
    series: Series
    patient: Patient
    equipment: Equipment
    optical_paths: Sequence[OpticalPath]
    slide: Slide
    label: Label
    image: Image
    frame_of_reference_uid: Optional[UID] = None
    dimension_organization_uids: Optional[Sequence[UID]] = None

    @cached_property
    def default_frame_of_reference_uid(self) -> UID:
        """Frame of reference uid used if not set."""
        if self.frame_of_reference_uid is not None:
            return self.frame_of_reference_uid
        return generate_uid()

    @cached_property
    def default_dimension_organization_uids(self) -> Sequence[UID]:
        """Dimension organization uids used if not set."""
        if self.dimension_organization_uids is not None:
            return self.dimension_organization_uids
        return [generate_uid()]

    @classmethod
    def merge_image_types(
        cls,
        volume: "WsiMetadata",
        label: Optional["WsiMetadata"],
        overview: Optional["WsiMetadata"],
    ):
        if volume.label is not None:
            if label is not None:
                label_label = label.label
            else:
                label_label = None
            if overview is not None:
                overview_label = overview.label
            else:
                overview_label = None

            merged_label = Label.merge_image_types(
                volume.label, label_label, overview_label
            )
        else:
            merged_label = Label()
        return replace(volume, label=merged_label)



#    Copyright 2023 SECTRA AB
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Optical path model."""

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import Generic, Optional, Sequence, Type, TypeVar, Union

import numpy as np

from wsidicom.conceptcode import (
    IlluminationCode,
    IlluminationColorCode,
    ImagePathFilterCode,
    LenseCode,
    LightPathFilterCode,
)

LutDataType = Type[Union[np.uint8, np.uint16]]


class LutSegment(metaclass=ABCMeta):
    """Metaclass for a LUT segment."""

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def array(self, data_type: LutDataType) -> np.ndarray:
        """Return the segment as an array of the given data type."""
        raise NotImplementedError()


@dataclass(frozen=True)
class DiscreteLutSegment(LutSegment):
    """A discrete segmented defined by a sequence of values.

    Note that when serializing, if the next segment is a Linear segment its start value
    will be included in this segment.

    Parameters
    ----------
    values: Sequence[int]
        The discrete values of the segment.
    """

    values: Sequence[int]

    def __len__(self) -> int:
        return len(self.values)

    def array(self, data_type: LutDataType) -> np.ndarray:
        return np.array(self.values, dtype=data_type)


@dataclass(frozen=True)
class LinearLutSegment(LutSegment):
    """A linear segment defined by start and end values and a length.

    Note that the corresponding Dicom Linear Segment Type does not define the start
    value. When serialized to Dicom the start value will be included in the previous
    segment (as last value in a Discrete or Linear segment) and the length reduced by
    one.

    Parameters
    ----------
    start_value: int
        The start value of the linear segment.
    end_value: int
        The end value of the linear segment.
    length: int
        The length of the linear segment.
    """

    start_value: int
    end_value: int
    length: int

    def __len__(self) -> int:
        return self.length

    def array(self, data_type: LutDataType) -> np.ndarray:
        return np.linspace(
            start=self.start_value,
            stop=self.end_value,
            num=self.length,
            dtype=data_type,
        )


@dataclass(frozen=True)
class ConstantLutSegment(LutSegment):
    """A constant segment defined by a value and a length.

    Parameters
    ----------
    value: int
        The value of the constant segment.
    length: int
        The length of the constant segment.
    """

    value: int
    length: int

    def __len__(self) -> int:
        return self.length

    def array(self, data_type: LutDataType) -> np.ndarray:
        return np.full(
            self.length,
            self.value,
            dtype=data_type,
        )


@dataclass(frozen=True)
class Lut:
    """Represents a LUT.

    Parameters
    ----------
    red: Sequence[LutSegment]
        The red segments of the LUT.
    green: Sequence[LutSegment]
        The green segments of the LUT.
    blue: Sequence[LutSegment]
        The blue segments of the LUT.
    data_type: LutDataType
        The data type of the LUT.
    """

    red: Sequence[LutSegment]
    green: Sequence[LutSegment]
    blue: Sequence[LutSegment]
    data_type: LutDataType

    @property
    def bits(self) -> int:
        if self.data_type == np.uint8:
            return 8
        if self.data_type == np.uint16:
            return 16
        raise ValueError("Value type should be 'np.uint8' or np.uint16'.")

    @property
    def table(self) -> np.ndarray:
        return np.stack(
            [
                np.concatenate([segment.array(self.data_type) for segment in self.red]),
                np.concatenate(
                    [segment.array(self.data_type) for segment in self.green]
                ),
                np.concatenate(
                    [segment.array(self.data_type) for segment in self.blue]
                ),
            ]
        )

    @property
    def length(self) -> int:
        length_by_color = set(
            sum(len(segment) for segment in color)
            for color in (self.red, self.green, self.blue)
        )
        if len(length_by_color) != 1:
            raise ValueError("Color components have different length.")
        return length_by_color.pop()

    def array(self, mode: str) -> np.ndarray:
        """Return flattened representation of the lookup table with order
        suitable for use with Pillows point(). The lookup table is scaled to
        either 8 or 16 bit depending on mode.

        Parameters
        ----------
        mode: str
            Image mode to produce lookup table for.

        Returns
        ----------
        np.ndarray
            Lookup table ordered by rgb, rgb ...
        """
        if mode == "L" or mode == "I":
            bits = 16
        else:
            bits = 8
        return self.table.flatten() / (2**self.bits / 2**bits)


OpticalFilterCodeType = TypeVar(
    "OpticalFilterCodeType",
    LightPathFilterCode,
    ImagePathFilterCode,
)
OpticalFilterType = TypeVar("OpticalFilterType", bound="OpticalFilter")


@dataclass(frozen=True)
class OpticalFilter(Generic[OpticalFilterCodeType]):
    """Metaclass for filter conditions for optical path.

    Parameters
    ----------
    filters: Sequence[OpticalFilterCodeType] = []
        The filters used.
    nominal: Optional[float] = None
        The nominal value of the filter in nm.
    low_pass: Optional[float] = None
        The low pass value of the filter in nm.
    high_pass: Optional[float] = None
        The high pass value of the filter in nm.
    """

    filters: Sequence[OpticalFilterCodeType] = field(default_factory=list)
    nominal: Optional[float] = None
    low_pass: Optional[float] = None
    high_pass: Optional[float] = None


class LightPathFilter(OpticalFilter[LightPathFilterCode]):
    """Set of light path filter conditions for optical path.

    Parameters
    ----------
    filters: Sequence[LightPathFilterCode] = []
        The filters used. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8124.html
        or `LightPathFilterCode.meanings` for a list of valid codes.
    nominal: Optional[float] = None
        The nominal value of the filter in nm.
    low_pass: Optional[float] = None
        The low pass value of the filter in nm.
    high_pass: Optional[float] = None
        The high pass value of the filter in nm.
    """


class ImagePathFilter(OpticalFilter[ImagePathFilterCode]):
    """Set of image path filter conditions for optical path.

    Parameters
    ----------
    filters: Sequence[ImagePathFilterCode] = []
        The filters used. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8124.html
        or ImagePathFilterCode.meanings` for a list of valid codes.
    nominal: Optional[float] = None
        The nominal value of the filter in nm.
    low_pass: Optional[float] = None
        The low pass value of the filter in nm.
    high_pass: Optional[float] = None
        The high pass value of the filter in nm.
    """


@dataclass(frozen=True)
class Objectives:
    """Set of lens conditions for optical path.

    Parameters
    ----------
    lenses: Sequence[LenseCode] = []
        Lenses used. See
        https://dicom.nema.org/medical/Dicom/current/output/chtml/part16/sect_CID_8121.html
        or `LensCode.meanings` for a list of valid codes.
    condenser_power: Optional[float] = None
        The condenser power.
    objective_power: Optional[float] = None
        The objective power.
    objective_numerical_aperture: Optional[float] = None
        The objective numerical aperture.
    """

    lenses: Sequence[LenseCode] = field(default_factory=list)
    condenser_power: Optional[float] = None
    objective_power: Optional[float] = None
    objective_numerical_aperture: Optional[float] = None


@dataclass(frozen=True)
class OpticalPath:
    """
    Optical path metadata.

    Corresponds to the `Required`, `Required, Empty if Unknown`, and selected
    `Optional` attributes for an Optical Path Sequence item in the Optical Path Module:
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.7.html

    Parameters
    identifier : Optional[str] = None
        Identifier of the optical path.
    description : Optional[str] = None
        Description of the optical path.
    illumination_types : Optional[Sequence[IlluminationCode]] = None
        Illumination types used in the optical path.
    illumination : Optional[Union[float, IlluminationColorCode]] = None
        Illumination used in the optical path.
    icc_profile : Optional[bytes] = None
        ICC profile for the optical path.
    lut : Optional[Lut] = None
        Lookup table to use for the optical path.
    light_path_filter : Optional[LightPathFilter] = None
        Light path filter used in the optical path.
    image_path_filter : Optional[ImagePathFilter] = None
        Image path filter used in the optical path.
    objective : Optional[Objectives] = None
        Objectives used in the optical path.
    """

    identifier: Optional[str] = None
    description: Optional[str] = None
    illumination_types: Optional[Sequence[IlluminationCode]] = None
    illumination: Optional[Union[float, IlluminationColorCode]] = None
    icc_profile: Optional[bytes] = None
    lut: Optional[Lut] = None
    light_path_filter: Optional[LightPathFilter] = None
    image_path_filter: Optional[ImagePathFilter] = None
    objective: Optional[Objectives] = None
